import json
import unittest
from unittest import mock

import hydrofunctions as hf


class TestFloodStage(unittest.TestCase):
    def test_correct_filtering(self):
        stages = {"07144100": {"action_stage": "20"}}
        expected = {"07144100": {"action_stage": "20"}, "07144101": None}
        stations = ["07144100", "07144101"]

        actual = hf.filter_flood_stages(stages, stations)
        self.assertDictEqual(actual, expected)

    @mock.patch("hydrofunctions.waterwatch._get_flood_stages")
    def test_waterwatch_get_stage(self, mock_get_fs):
        site = "07144100"
        expected = {
            site: {
                "action_stage": "20",
                "flood_stage": "22",
                "moderate_flood_stage": "25",
            }
        }
        mock_get_fs.content = json.dumps(
            {
                "sites": [
                    {
                        "site_no": site,
                        "action_stage": "20",
                        "flood_stage": "22",
                        "moderate_flood_stage": "25",
                    }
                ]
            },
            separators=(",", ":"),
        ).encode("utf-8")
        mock_get_fs.return_value = expected

        actual = hf.get_flood_stage(site=site)
        self.assertEqual(actual, expected)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-
"""
Created on Sun Feb 25 20:07:38 2018

@author: Marty
"""

from __future__ import absolute_import, print_function, division, unicode_literals
import unittest
from IPython.core.display import HTML
import re

import hydrofunctions as hf


class TestMap(unittest.TestCase):
    def test_helpers_draw_map_returns_HTML_obj(self):
        actual = hf.draw_map()
        self.assertIsInstance(actual, HTML)

    def test_helpers_draw_map_defaults(self):
        # Search through the generated html for the default values.
        widthRE = re.compile(r"width=700")
        heightRE = re.compile(r"height=400")
        urlRE = re.compile(r"http://hydrocloud.org")

        actual = hf.draw_map()
        self.assertTrue(widthRE.search(actual.data))
        self.assertTrue(heightRE.search(actual.data))
        self.assertTrue(urlRE.search(actual.data))

    def test_helpers_count_truthy_two_true(self):
        expect_two1 = [False, True, False, True]
        self.assertEqual(hf.count_number_of_truthy(expect_two1), 2, "test 1 failed.")

        expect_two2 = ["text", False, False, True]
        self.assertEqual(hf.count_number_of_truthy(expect_two2), 2, "test 2 failed.")

        expect_two3 = [True, True]
        self.assertEqual(hf.count_number_of_truthy(expect_two3), 2, "test 3 failed.")

        expect_two4 = [5, "text", False, False]
        self.assertEqual(hf.count_number_of_truthy(expect_two4), 2, "test 4 failed.")

    def test_helpers_count_truthy_no_true(self):
        expect_no_true1 = [False, False, False, False]
        self.assertEqual(
            hf.count_number_of_truthy(expect_no_true1), 0, "test 1 failed."
        )

        expect_no_true2 = [None, False, False, None]
        self.assertEqual(
            hf.count_number_of_truthy(expect_no_true2), 0, "test 2 failed."
        )

        expect_no_true3 = []
        self.assertEqual(
            hf.count_number_of_truthy(expect_no_true3), 0, "test 3 failed."
        )



# -*- coding: utf-8 -*-
"""
fixtures.py

This module is for storing all of the relavant fixtures used in testing.
"""
from .fixtures_data import (
    JSON15min2day,
    two_sites_two_params_iv,
    nothing_avail,
    mult_flags,
    diff_freq,
    startDST,
    endDST,
    daily_lake_level,
)

from .fixtures_daily_dupe import daily_dupe, daily_dupe_altered

from .fixtures_multiple_methods import multi_meth

from .fixtures_tzfail import tzfail

from .fixtures_recent_only import recent_only

from .fixtures_usgs_rdb import (
    field_fixture,
    rating_fixture,
    peaks_fixture,
    parsing_error_fixture,
)


class fakeResponse(object):
    def __init__(
        self,
        code=200,
        url="fake url",
        reason="fake reason",
        text="fake text",
        json=JSON15min2day,
    ):
        self.status_code = code
        self.url = url
        self.reason = reason
        self.text = text
        # .json will return a function
        # .json() will return JSON15min2day
        self.json = lambda: json
        if self.status_code == 200:
            self.ok = True
        else:
            self.ok = False

    def raise_for_status(self):
        return self.status_code



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_hydrofunctions
----------------------------------

Tests for `hydrofunctions` module.
"""
from __future__ import (
    absolute_import,
    print_function,
    division,
    unicode_literals,
)
from unittest import mock
import unittest
import warnings
import os

from pandas.testing import assert_frame_equal

import pandas as pd
import numpy as np
import pyarrow as pa
import json

from hydrofunctions.exceptions import HydroNoDataError

print("Pyarrow version: ", pa.__version__)

import hydrofunctions as hf
from .fixtures import (
    fakeResponse,
    daily_dupe,
    daily_dupe_altered,
    multi_meth,
    tzfail,
    JSON15min2day,
    two_sites_two_params_iv,
    nothing_avail,
    mult_flags,
    diff_freq,
    startDST,
    endDST,
    recent_only,
)


class TestHydrofunctionsParsing(unittest.TestCase):
    """Test the parsing of hf.extract_nwis_df()

    test the following:
        Can it handle multiple qualifier flags?
        how does it encode mult params & mult sites?
        Does it raise HydroNoDataError if nothing returned?

    """

    def test_hf_extract_nwis_df_accepts_response_obj(self):
        fake_response = fakeResponse()
        actual_df, actual_dict = hf.extract_nwis_df(fake_response, interpolate=False)
        self.assertIsInstance(
            actual_df, pd.core.frame.DataFrame, msg="Did not return a df"
        )
        self.assertIsInstance(actual_dict, dict, msg="Did not return a dict.")

    def test_hf_extract_nwis_df_parse_multiple_flags(self):
        actual_df, actual_dict = hf.extract_nwis_df(mult_flags, interpolate=False)
        self.assertIsInstance(
            actual_df, pd.core.frame.DataFrame, msg="Did not return a df"
        )
        self.assertIsInstance(actual_dict, dict, msg="Did not return a dict.")

    def test_hf_extract_nwis_df_parse_two_sites_two_params_iv_return_meta_dict(self):
        actual_df, actual_dict = hf.extract_nwis_df(
            two_sites_two_params_iv, interpolate=False
        )
        self.assertIs(
            type(actual_df), pd.core.frame.DataFrame, msg="Did not return a df"
        )
        self.assertIs(type(actual_dict), dict, msg="Did not return a dict.")
        # TODO: test that metadata is organized correctly

    def test_hf_extract_nwis_df_parse_two_sites_two_params_iv_cols(self):
        actual_df, actual_dict = hf.extract_nwis_df(
            two_sites_two_params_iv, interpolate=False
        )
        actual_len, actual_width = actual_df.shape
        self.assertIs(
            type(actual_df), pd.core.frame.DataFrame, msg="Did not return a df"
        )
        self.assertEqual(actual_len, 93, "Wrong length for dataframe")
        self.assertEqual(actual_width, 8, "Wrong width for dataframe")
        expected_columns = [
            "USGS:01541000:00060:00000",
            "USGS:01541000:00060:00000_qualifiers",
            "USGS:01541000:00065:00000",
            "USGS:01541000:00065:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541200:00065:00000_qualifiers",
        ]
        actual_columns = actual_df.columns.values
        self.assertCountEqual(
            actual_columns, expected_columns, "column names don't match expected"
        )
        self.assertTrue(actual_df.index.is_unique, "index has repeated values.")
        self.assertTrue(actual_df.index.is_monotonic, "index is not monotonic.")

    def test_hf_extract_nwis_df_parse_JSON15min2day_return_df(self):
        actual_df, actual_dict = hf.extract_nwis_df(JSON15min2day, interpolate=False)
        actual_len, actual_width = actual_df.shape
        self.assertIs(
            type(actual_df), pd.core.frame.DataFrame, msg="Did not return a df"
        )
        self.assertEqual(actual_len, 192, "Wrong length for dataframe")
        self.assertEqual(actual_width, 2, "Wrong width for dataframe")
        expected_columns = [
            "USGS:03213700:00060:00000",
            "USGS:03213700:00060:00000_qualifiers",
        ]
        actual_columns = actual_df.columns.values
        self.assertCountEqual(
            actual_columns, expected_columns, "column names don't match expected"
        )
        self.assertTrue(actual_df.index.is_unique, "index has repeated values.")
        self.assertTrue(actual_df.index.is_monotonic, "index is not monotonic.")

    def test_hf_extract_nwis_df_parse_mult_flags_return_df(self):
        actual_df, actual_dict = hf.extract_nwis_df(mult_flags, interpolate=False)
        actual_len, actual_width = actual_df.shape
        self.assertIs(
            type(actual_df), pd.core.frame.DataFrame, msg="Did not return a df"
        )
        self.assertEqual(actual_len, 480, "Wrong length for dataframe")
        self.assertEqual(actual_width, 2, "Wrong width for dataframe")
        expected_columns = [
            "USGS:01542500:00060:00000",
            "USGS:01542500:00060:00000_qualifiers",
        ]
        actual_columns = actual_df.columns.values
        self.assertCountEqual(
            actual_columns, expected_columns, "column names don't match expected"
        )
        self.assertTrue(actual_df.index.is_unique, "index has repeated values.")
        self.assertTrue(actual_df.index.is_monotonic, "index is not monotonic.")

    def test_hf_extract_nwis_raises_exception_when_df_is_empty(self):
        empty_response = {"value": {"timeSeries": []}}
        with self.assertRaises(hf.HydroNoDataError):
            hf.extract_nwis_df(empty_response, interpolate=False)

    def test_hf_extract_nwis_raises_exception_when_df_is_empty_nothing_avail(self):
        with self.assertRaises(hf.HydroNoDataError):
            hf.extract_nwis_df(nothing_avail, interpolate=False)

    @unittest.skip(
        "assertWarns errors on Linux. See https://bugs.python.org/issue29620"
    )
    def test_hf_extract_nwis_warns_when_diff_series_have_diff_freq(self):
        with self.assertWarns(hf.HydroUserWarning):
            hf.extract_nwis_df(diff_freq, interpolate=False)

    def test_hf_extract_nwis_accepts_no_startdate_no_period_no_interpolate(self):
        actual_df, actual_dict = hf.extract_nwis_df(recent_only, interpolate=False)
        expected_shape = (
            2,
            4,
        )  # only the most recent data for two parameters, plus qualifiers = 4 columns; 2 rows: different dates.
        self.assertEqual(
            actual_df.shape,
            expected_shape,
            "The dataframe should have four columns and two rows.",
        )

    def test_hf_extract_nwis_accepts_no_startdate_no_period_interpolate(self):
        actual_df, actual_dict = hf.extract_nwis_df(recent_only, interpolate=True)
        expected_shape = (
            2,
            4,
        )  #  only the most recent data for two parameters, plus qualifiers = 4 columns; 2 rows: different dates.
        self.assertEqual(
            actual_df.shape,
            expected_shape,
            "The dataframe should have four columns and two rows.",
        )

    def test_hf_extract_nwis_returns_comma_separated_qualifiers_1(self):
        actual_df, actual_dict = hf.extract_nwis_df(mult_flags, interpolate=False)
        actual_flags_1 = actual_df.loc[
            "2019-01-24T10:30:00.000-05:00", "USGS:01542500:00060:00000_qualifiers"
        ]
        expected_flags_1 = "P,e"
        self.assertEqual(
            actual_flags_1,
            expected_flags_1,
            "The data qualifier flags were not parsed correctly.",
        )

    def test_hf_extract_nwis_returns_comma_separated_qualifiers_2(self):
        actual_df, actual_dict = hf.extract_nwis_df(mult_flags, interpolate=False)
        actual_flags_2 = actual_df.loc[
            "2019-01-28T16:00:00.000-05:00", "USGS:01542500:00060:00000_qualifiers"
        ]
        expected_flags_2 = "P,Ice"
        self.assertEqual(
            actual_flags_2,
            expected_flags_2,
            "The data qualifier flags were not parsed correctly.",
        )

    def test_hf_extract_nwis_replaces_NWIS_noDataValue_with_npNan(self):
        actual_df, actual_dict = hf.extract_nwis_df(mult_flags, interpolate=False)
        actual_nodata = actual_df.loc[
            "2019-01-28T16:00:00.000-05:00", "USGS:01542500:00060:00000"
        ]
        self.assertTrue(
            np.isnan(actual_nodata),
            "The NWIS no data value was not replaced with np.nan. ",
        )

    def test_hf_extract_nwis_adds_missing_tags(self):
        actual_df, actual_dict = hf.extract_nwis_df(mult_flags, interpolate=False)
        actual_missing = actual_df.loc[
            "2019-01-24 17:00:00-05:00", "USGS:01542500:00060:00000_qualifiers"
        ]
        self.assertEqual(
            actual_missing,
            "hf.missing",
            "Missing records should be given 'hf.missing' _qualifier tags.",
        )

    def test_hf_extract_nwis_adds_upsample_tags(self):
        actual_df, actual_dict = hf.extract_nwis_df(diff_freq, interpolate=False)
        actual_upsample = actual_df.loc[
            "2018-06-01 00:15:00-04:00", "USGS:01570500:00060:00000_qualifiers"
        ]
        self.assertEqual(
            actual_upsample,
            "hf.upsampled",
            "New records created by upsampling should be given 'hf.upsample' _qualifier tags.",
        )

    def test_hf_extract_nwis_interpolates(self):
        actual_df, actual_dict = hf.extract_nwis_df(diff_freq, interpolate=True)
        actual_upsample_interpolate = actual_df.loc[
            "2018-06-01 00:15:00-04:00", "USGS:01570500:00060:00000"
        ]
        self.assertEqual(
            actual_upsample_interpolate,
            42200.0,
            "New records created by upsampling should have NaNs replaced with interpolated values.",
        )

    @unittest.skip("This feature is not implemented yet.")
    def test_hf_extract_nwis_interpolates_and_adds_tags(self):
        # Ideally, every data value that was interpolated should have a tag
        # added to the qualifiers that says it was interpolated.
        actual_df, actual_dict = hf.extract_nwis_df(diff_freq, interpolate=True)
        actual_upsample_interpolate_flag = actual_df.loc[
            "2018-06-01 00:15:00-04:00", "USGS:01570500:00060:00000_qualifiers"
        ]
        expected_flag = "hf.interpolated"
        self.assertEqual(
            actual_upsample_interpolate_flag,
            expected_flag,
            "Interpolated values should be marked with a flag.",
        )

    def test_hf_extract_nwis_corrects_for_start_of_DST(self):
        actual_df, actual_dict = hf.extract_nwis_df(startDST, interpolate=False)
        actual_len, width = actual_df.shape
        expected = 284
        self.assertEqual(
            actual_len,
            expected,
            "Three days including the start of DST should have 3 * 24 * 4 = 288 observations, minus 4 = 284",
        )

    def test_hf_extract_nwis_corrects_for_end_of_DST(self):
        actual_df, actual_dict = hf.extract_nwis_df(endDST, interpolate=False)
        actual_len, width = actual_df.shape
        expected = 292
        self.assertEqual(
            actual_len,
            expected,
            "Three days including the end of DST should have 3 * 24 * 4 = 288 observations, plus 4 = 292",
        )

    def test_hf_extract_nwis_can_find_tz_in_tzfail(self):
        actualDF = hf.extract_nwis_df(tzfail, interpolate=False)

    def test_hf_extract_nwis_can_deal_with_duplicated_records_as_input(self):
        actualDF = hf.extract_nwis_df(daily_dupe, interpolate=False)

    def test_hf_extract_nwis_can_deal_with_duplicated_records_that_have_been_altered_as_input(
        self,
    ):
        # What happens if a scientist replaces an empty record with new
        # estimated data, and forgets to discard the old data?
        actualDF = hf.extract_nwis_df(daily_dupe_altered, interpolate=False)

    def test_hf_extract_nwis_multiple_methods_one_param(self):
        actual_df, actual_meta = hf.extract_nwis_df(multi_meth, interpolate=False)
        actual_len, actual_width = actual_df.shape
        # Only one parameter, but if the json is processed properly, each
        # method will be given a column and a qualifier column.
        expected_width = 4
        self.assertEqual(
            actual_width,
            expected_width,
            "Extract_nwis should have parsed two methods into separate columns.",
        )

    def test_hf_get_nwis_property(self):
        sites = None
        bBox = (-105.430, 39.655, -104, 39.863)
        # TODO: test should be the json for a multiple site request.
        names = hf.get_nwis_property(JSON15min2day, key="name")
        self.assertIs(type(names), list, msg="Did not return a list")


class TestHydrofunctions(unittest.TestCase):
    @mock.patch("requests.get")
    def test_hf_get_nwis_calls_correct_url(self, mock_get):

        """
        Thanks to
        http://engineroom.trackmaven.com/blog/making-a-mockery-of-python/
        """

        site = "A"
        service = "iv"
        start = "C"
        end = "D"

        expected_url = "https://waterservices.usgs.gov/nwis/" + service + "/?"
        expected_headers = {"Accept-encoding": "gzip", "max-age": "120"}
        expected_params = {
            "format": "json,1.1",
            "sites": "A",
            "stateCd": None,
            "countyCd": None,
            "bBox": None,
            "parameterCd": None,
            "period": None,
            "startDT": "C",
            "endDT": "D",
        }

        expected = fakeResponse()
        expected.status_code = 200
        expected.reason = "any text"

        mock_get.return_value = expected
        actual = hf.get_nwis(site, service, start, end)
        mock_get.assert_called_once_with(
            expected_url, params=expected_params, headers=expected_headers
        )
        self.assertEqual(actual, expected)

    @mock.patch("requests.get")
    def test_hf_get_nwis_calls_correct_url_multiple_sites(self, mock_get):

        site = ["site1", "site2"]
        parsed_site = hf.check_parameter_string(site, "site")
        service = "iv"
        start = "C"
        end = "D"

        expected_url = "https://waterservices.usgs.gov/nwis/" + service + "/?"
        expected_headers = {"max-age": "120", "Accept-encoding": "gzip"}
        expected_params = {
            "format": "json,1.1",
            "sites": parsed_site,
            "stateCd": None,
            "countyCd": None,
            "bBox": None,
            "parameterCd": None,
            "period": None,
            "startDT": "C",
            "endDT": "D",
        }

        expected = fakeResponse()
        expected.status_code = 200
        expected.reason = "any text"

        mock_get.return_value = expected
        actual = hf.get_nwis(site, service, start, end)
        mock_get.assert_called_once_with(
            expected_url, params=expected_params, headers=expected_headers
        )
        self.assertEqual(actual, expected)

    @mock.patch("requests.get")
    def test_hf_get_nwis_service_defaults_dv(self, mock_get):
        site = "01541200"
        expected_service = "dv"

        expected_url = "https://waterservices.usgs.gov/nwis/" + expected_service + "/?"
        expected_headers = {"max-age": "120", "Accept-encoding": "gzip"}
        expected_params = {
            "format": "json,1.1",
            "sites": site,
            "stateCd": None,
            "countyCd": None,
            "bBox": None,
            "parameterCd": None,
            "period": None,
            "startDT": None,
            "endDT": None,
        }
        expected = fakeResponse()
        expected.status_code = 200
        expected.reason = "any text"

        mock_get.return_value = expected
        actual = hf.get_nwis(site)
        mock_get.assert_called_once_with(
            expected_url, params=expected_params, headers=expected_headers
        )
        self.assertEqual(actual, expected)

    @mock.patch("requests.get")
    def test_hf_get_nwis_converts_parameterCd_all_to_None(self, mock_get):
        site = "01541200"
        service = "iv"
        parameterCd = "all"
        expected_parameterCd = None
        expected_url = "https://waterservices.usgs.gov/nwis/" + service + "/?"
        expected_headers = {"max-age": "120", "Accept-encoding": "gzip"}
        expected_params = {
            "format": "json,1.1",
            "sites": site,
            "stateCd": None,
            "countyCd": None,
            "bBox": None,
            "parameterCd": None,
            "period": None,
            "startDT": None,
            "endDT": None,
        }
        expected = fakeResponse()
        expected.status_code = 200
        expected.reason = "any text"

        mock_get.return_value = expected
        actual = hf.get_nwis(site, service, parameterCd=parameterCd)
        mock_get.assert_called_once_with(
            expected_url, params=expected_params, headers=expected_headers
        )
        self.assertEqual(actual, expected)

    def test_hf_get_nwis_raises_ValueError_too_many_locations(self):
        with self.assertRaises(ValueError):
            hf.get_nwis("01541000", stateCd="MD")

    def test_hf_get_nwis_raises_ValueError_start_and_period(self):
        with self.assertRaises(ValueError):
            hf.get_nwis("01541000", start_date="2014-01-01", period="P1D")

    @mock.patch("requests.get")
    @mock.patch("builtins.print")
    def test_hf_get_nwis_verbose_True(self, mock_print, mock_get):
        # testing a print function from https://realpython.com/lessons/mocking-print-unit-tests/
        expected = fakeResponse()
        expected.status_code = 200
        expected.reason = "any text"
        expected.url = "expected url"
        mock_get.return_value = expected
        expected_text = "Requested data from"

        actual = hf.get_nwis("01582500", period="P2D", verbose=True)

        mock_print.assert_called_with(expected_text, expected.url)

    @mock.patch("requests.get")
    @mock.patch("builtins.print")
    def test_hf_get_nwis_verbose_False(self, mock_print, mock_get):
        # testing a print function from https://realpython.com/lessons/mocking-print-unit-tests/
        expected = fakeResponse()
        expected.status_code = 200
        expected.reason = "any text"
        mock_get.return_value = expected

        actual = hf.get_nwis("01582500", period="P2D", verbose=False)

        mock_print.assert_not_called()

    def test_hf_nwis_custom_status_codes_returns_None_for_200(self):
        fake = fakeResponse()
        fake.status_code = 200
        fake.reason = "any text"
        fake.url = "any text"
        self.assertIsNone(hf.nwis_custom_status_codes(fake))

    def test_hf_nwis_custom_status_codes_raises_HydroNoDataError_for_non200(self):
        expected_status_code = 400
        bad_response = fakeResponse(code=expected_status_code)
        with self.assertRaises(
            HydroNoDataError, msg="This should raise a HydroNoDataError."
        ) as cm:
            hf.nwis_custom_status_codes(bad_response)

    def test_hf_calc_freq_returns_Timedelta_and_60min(self):
        test_index = pd.date_range("2014-12-29", "2015-01-03", freq="60T")
        actual = hf.calc_freq(test_index)
        self.assertIsInstance(
            actual, pd.Timedelta, "Calc_freq() should return pd.Timedelta."
        )
        expected = pd.Timedelta("60 minutes")
        self.assertEqual(
            actual, expected, "Calc_freq() should have converted 60T to 60 minutes."
        )

    def test_hf_calc_freq_accepts_Day(self):
        test_index = pd.date_range("2014-12-29", periods=3)
        actual = hf.calc_freq(test_index)
        self.assertIsInstance(
            actual, pd.Timedelta, "Calc_freq() should return pd.Timedelta."
        )
        expected = pd.Timedelta("1 day")
        self.assertEqual(
            actual, expected, "Calc_freq() should have found a 1 day frequency."
        )

    def test_hf_calc_freq_accepts_hour(self):
        test_index = pd.date_range("2014-12-29", freq="1H", periods=30)
        actual = hf.calc_freq(test_index)
        self.assertIsInstance(
            actual, pd.Timedelta, "Calc_freq() should return pd.Timedelta."
        )
        expected = pd.Timedelta("1 hour")
        self.assertEqual(
            actual, expected, "Calc_freq() should have found a 1 hour frequency."
        )

    def test_hf_calc_freq_accepts_1Day_1hour(self):
        test_index = pd.date_range("2014-12-29", freq="1D1H2T", periods=30)
        actual = hf.calc_freq(test_index)
        self.assertIsInstance(
            actual, pd.Timedelta, "Calc_freq() should return pd.Timedelta."
        )
        expected = pd.Timedelta("1 day 1 hour 2 minutes")
        self.assertEqual(
            actual,
            expected,
            "Calc_freq() should have found a 1 day, 1 hour, 2 minutes frequency.",
        )

    def test_hf_calc_freq_accepts_freq_None(self):
        dates = ["2014-12-20", "2014-12-22", "2014-12-24", "2014-12-26"]
        test_index = pd.DatetimeIndex(dates)
        self.assertIsNone(
            test_index.freq,
            msg="The test_index was not properly set up so that test_index.freq is None.",
        )
        actual = hf.calc_freq(test_index)
        self.assertIsInstance(
            actual, pd.Timedelta, "Calc_freq() should return pd.Timedelta."
        )
        expected = pd.Timedelta("48 hours")
        self.assertEqual(
            actual, expected, "Calc_freq() should have returned a 48 hour period."
        )

    def test_hf_calc_freq_accepts_df(self):
        test_index = pd.date_range("2014-12-29", "2014-12-30", freq="1T")
        test_df = pd.DataFrame(index=test_index)
        actual = hf.calc_freq(test_df)
        self.assertIsInstance(
            actual, pd.Timedelta, "Calc_freq() should return pd.Timedelta."
        )
        expected = pd.Timedelta("1 minute")
        self.assertEqual(
            actual, expected, "Calc_freq() should have returned a 1 minute period."
        )

    def test_hf_calc_freq_accepts_difficult_ts_freq_deleted(self):
        test_index = pd.date_range("2014-12-29", "2014-12-30", freq="1H")
        actual = hf.calc_freq(test_index)
        self.assertIsInstance(
            actual, pd.Timedelta, "Calc_freq() should return pd.Timedelta."
        )
        expected = pd.Timedelta("1 hour")
        self.assertEqual(
            actual, expected, "Calc_freq() should have returned a 1 hour frequency."
        )

    @unittest.skip("Difficulty dropping rows.")
    def test_hf_calc_freq_accepts_difficult_ts_freq_deleted_row_dropped(self):
        test_index = pd.date_range("2014-12-29", "2014-12-30", freq="1H")
        test_df = pd.DataFrame(index=test_index)
        # Can't get this to work.
        test_df.drop("2014-12-30", axis=0)
        actual = hf.calc_freq(test_df)
        self.assertIsInstance(
            actual, pd.Timedelta, "Calc_freq() should return pd.Timedelta."
        )
        expected = pd.Timedelta("1 hour")
        self.assertEqual(
            actual, expected, "Calc_freq() should have returned a 1 hour frequency."
        )

    @unittest.skip("")
    def test_hf_calc_freq_accepts_difficult_ts_freq_deleted_no_rows(self):
        pass

    @unittest.skip("")
    def test_hf_calc_freq_accepts_difficult_ts_freq_deleted_no_time_index(self):
        pass

    @unittest.skip("Not sure how to trigger a warning.")
    def test_hf_calc_freq_raises_warning(self):
        test_df = pd.DataFrame(data={"A": [1, 2, 3], "B": [4, 5, 6]})
        actual = hf.calc_freq(test_df)
        self.assertIsInstance(
            actual, pd.Timedelta, "Calc_freq() should return pd.Timedelta."
        )
        expected = pd.Timedelta("15 minutes")
        self.assertEqual(
            actual, expected, "Calc_freq() should have returned a 15 minute frequency."
        )

    def test_hf_select_data_returns_data_cols(self):
        actual_df, actual_dict = hf.extract_nwis_df(two_sites_two_params_iv)
        actual_df = actual_df.reindex(sorted(actual_df.columns), axis=1)
        actual = hf.select_data(actual_df)
        expected = [True, False, True, False, True, False, True, False]
        self.assertListEqual(
            actual.tolist(),
            expected,
            "select_data should return an array of which columns contain the data, not the qualifiers.",
        )

    def test_hf_save_read_parquet_integration(self):
        # This test has side effects: it will create a file.
        expected_df, expected_meta = hf.extract_nwis_df(two_sites_two_params_iv)
        filename = "test_filename_delete_me.parquet"
        hf.save_parquet(filename, expected_df, expected_meta)
        actual_df, actual_meta = hf.read_parquet(filename)
        os.remove("test_filename_delete_me.parquet")
        self.assertEqual(
            expected_df.index.freq,
            actual_df.index.freq,
            "The saved dataframe does not have its index frequency set correctly.",
        )
        assert_frame_equal(expected_df, actual_df)
        self.assertEqual(expected_meta, actual_meta, "The metadata dict has changed.")

    @mock.patch("pyarrow.parquet.read_table")
    def test_hf_read_parquet(self, mock_read):
        expected_df, expected_meta = hf.extract_nwis_df(two_sites_two_params_iv)
        expected_table = pa.Table.from_pandas(expected_df)
        meta_dict = expected_table.schema.metadata
        meta_string = json.dumps(expected_meta).encode()
        meta_dict[b"hydrofunctions_meta"] = meta_string
        expected_table = expected_table.replace_schema_metadata(meta_dict)
        mock_read.return_value = expected_table
        actual_df, actual_meta = hf.read_parquet("fake_filename.parquet")

        assert_frame_equal(expected_df, actual_df)
        self.assertEqual(expected_meta, actual_meta, "The metadata dict has changed.")

    @mock.patch("pyarrow.parquet.write_table")
    def test_hf_save_parquet(self, mock_write):
        filename = "expected_filename.parquet"
        expected_df, expected_meta = hf.extract_nwis_df(two_sites_two_params_iv)
        hf.save_parquet(filename, expected_df, expected_meta)

    @mock.patch("gzip.open")
    @mock.patch("json.loads")
    def test_hf_read_json_gz(self, mock_json, mock_gzip):
        expected = two_sites_two_params_iv
        mock_json.return_value = expected
        actual = hf.read_json_gzip("filename.json.gz")
        mock_gzip.assert_called_with("filename.json.gz", "rb")
        self.assertEqual(actual, expected)

    @mock.patch("gzip.open")
    @mock.patch("json.dump")
    def test_hf_save_json_gz(self, mock_json, mock_gzip):
        expected = two_sites_two_params_iv
        expected_filename = "save.json.gz"
        hf.save_json_gzip(expected_filename, expected)
        mock_gzip.assert_called_with(expected_filename, "wt", encoding="ascii")
        mock_json.assert_called()


if __name__ == "__main__":
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-
"""
Created on Wed Mar 20 14:13:44 2019

@author: Marty
"""

# This json is from a site that has a duplicated record.
# Also, it has an interesting flag, "Dis", for discontinued.
# hf.NWIS('03107698', "dv", '2017-04-23', '2017-04-29')

daily_dupe = {
    "name": "ns1:timeSeriesResponseType",
    "declaredType": "org.cuahsi.waterml.TimeSeriesResponseType",
    "scope": "javax.xml.bind.JAXBElement$GlobalScope",
    "value": {
        "queryInfo": {
            "queryURL": "http://waterservices.usgs.gov/nwis/dv/format=json%2C1.1&sites=03107698&startDT=2017-04-23&endDT=2017-04-29",
            "criteria": {
                "locationParam": "[ALL:03107698]",
                "variableParam": "ALL",
                "timeParam": {
                    "beginDateTime": "2017-04-23T00:00:00.000",
                    "endDateTime": "2017-04-29T00:00:00.000",
                },
                "parameter": [],
            },
            "note": [
                {"value": "[ALL:03107698]", "title": "filter:sites"},
                {
                    "value": "[mode=RANGE, modifiedSince=null] interval={INTERVAL[2017-04-23T00:00:00.000-04:00/2017-04-29T00:00:00.000-04:00]}",
                    "title": "filter:timeRange",
                },
                {"value": "methodIds=[ALL]", "title": "filter:methodId"},
                {"value": "2019-03-20T18:19:45.800Z", "title": "requestDT"},
                {"value": "bd536170-4b3c-11e9-9463-6cae8b6642f6", "title": "requestId"},
                {
                    "value": "Provisional data are subject to revision. Go to http://waterdata.usgs.gov/nwis/help/?provisional for more information.",
                    "title": "disclaimer",
                },
                {"value": "caas01", "title": "server"},
            ],
        },
        "timeSeries": [
            {
                "sourceInfo": {
                    "siteName": "Traverse Creek near Kendall, PA",
                    "siteCode": [
                        {"value": "03107698", "network": "NWIS", "agencyCode": "USGS"}
                    ],
                    "timeZoneInfo": {
                        "defaultTimeZone": {
                            "zoneOffset": "-05:00",
                            "zoneAbbreviation": "EST",
                        },
                        "daylightSavingsTimeZone": {
                            "zoneOffset": "-04:00",
                            "zoneAbbreviation": "EDT",
                        },
                        "siteUsesDaylightSavingsTime": True,
                    },
                    "geoLocation": {
                        "geogLocation": {
                            "srs": "EPSG:4326",
                            "latitude": 40.5261111,
                            "longitude": -80.45138889,
                        },
                        "localSiteXY": [],
                    },
                    "note": [],
                    "siteType": [],
                    "siteProperty": [
                        {"value": "ST", "name": "siteTypeCd"},
                        {"value": "05030101", "name": "hucCd"},
                        {"value": "42", "name": "stateCd"},
                        {"value": "42007", "name": "countyCd"},
                    ],
                },
                "variable": {
                    "variableCode": [
                        {
                            "value": "00060",
                            "network": "NWIS",
                            "vocabulary": "NWIS:UnitValues",
                            "variableID": 45807197,
                            "default": True,
                        }
                    ],
                    "variableName": "Streamflow, ft&#179;/s",
                    "variableDescription": "Discharge, cubic feet per second",
                    "valueType": "Derived Value",
                    "unit": {"unitCode": "ft3/s"},
                    "options": {
                        "option": [
                            {
                                "value": "Mean",
                                "name": "Statistic",
                                "optionCode": "00003",
                            }
                        ]
                    },
                    "note": [],
                    "noDataValue": -999999.0,
                    "variableProperty": [],
                    "oid": "45807197",
                },
                "values": [
                    {
                        "value": [
                            {
                                "value": "6.82",
                                "qualifiers": ["A"],
                                "dateTime": "2017-04-23T00:00:00.000",
                            },
                            {
                                "value": "6.04",
                                "qualifiers": ["A"],
                                "dateTime": "2017-04-24T00:00:00.000",
                            },
                            {
                                "value": "5.62",
                                "qualifiers": ["A"],
                                "dateTime": "2017-04-25T00:00:00.000",
                            },
                            {
                                "value": "-999999",
                                "qualifiers": ["P", "Dis"],
                                "dateTime": "2017-04-26T00:00:00.000",
                            },
                            {
                                "value": "-999999",
                                "qualifiers": ["P", "Dis"],
                                "dateTime": "2017-04-26T00:00:00.000",
                            },
                            {
                                "value": "-999999",
                                "qualifiers": ["P", "Dis"],
                                "dateTime": "2017-04-27T00:00:00.000",
                            },
                            {
                                "value": "-999999",
                                "qualifiers": ["P", "Dis"],
                                "dateTime": "2017-04-28T00:00:00.000",
                            },
                            {
                                "value": "-999999",
                                "qualifiers": ["P", "Dis"],
                                "dateTime": "2017-04-29T00:00:00.000",
                            },
                        ],
                        "qualifier": [
                            {
                                "qualifierCode": "Dis",
                                "qualifierDescription": "Record has been discontinued at the measurement site.",
                                "qualifierID": 0,
                                "network": "NWIS",
                                "vocabulary": "uv_rmk_cd",
                            },
                            {
                                "qualifierCode": "A",
                                "qualifierDescription": "Approved for publication -- Processing and review completed.",
                                "qualifierID": 1,
                                "network": "NWIS",
                                "vocabulary": "uv_rmk_cd",
                            },
                            {
                                "qualifierCode": "P",
                                "qualifierDescription": "Provisional data subject to revision.",
                                "qualifierID": 2,
                                "network": "NWIS",
                                "vocabulary": "uv_rmk_cd",
                            },
                        ],
                        "qualityControlLevel": [],
                        "method": [{"methodDescription": "", "methodID": 120332}],
                        "source": [],
                        "offset": [],
                        "sample": [],
                        "censorCode": [],
                    }
                ],
                "name": "USGS:03107698:00060:00003",
            }
        ],
    },
    "nil": False,
    "globalScope": True,
    "typeSubstituted": False,
}

# What happens if a scientist replaces an empty record with new
# estimated data, and forgets to discard the old data?
# This fixture simulates a substituted record that has been marked with 'e'
# in the qualifier tags. An entry has also been made in the qualifierCode
# list, but I had to fake that.
daily_dupe_altered = {
    "name": "ns1:timeSeriesResponseType",
    "declaredType": "org.cuahsi.waterml.TimeSeriesResponseType",
    "scope": "javax.xml.bind.JAXBElement$GlobalScope",
    "value": {
        "queryInfo": {
            "queryURL": "http://waterservices.usgs.gov/nwis/dv/format=json%2C1.1&sites=03107698&startDT=2017-04-23&endDT=2017-04-29",
            "criteria": {
                "locationParam": "[ALL:03107698]",
                "variableParam": "ALL",
                "timeParam": {
                    "beginDateTime": "2017-04-23T00:00:00.000",
                    "endDateTime": "2017-04-29T00:00:00.000",
                },
                "parameter": [],
            },
            "note": [
                {"value": "[ALL:03107698]", "title": "filter:sites"},
                {
                    "value": "[mode=RANGE, modifiedSince=null] interval={INTERVAL[2017-04-23T00:00:00.000-04:00/2017-04-29T00:00:00.000-04:00]}",
                    "title": "filter:timeRange",
                },
                {"value": "methodIds=[ALL]", "title": "filter:methodId"},
                {"value": "2019-03-20T18:19:45.800Z", "title": "requestDT"},
                {"value": "bd536170-4b3c-11e9-9463-6cae8b6642f6", "title": "requestId"},
                {
                    "value": "Provisional data are subject to revision. Go to http://waterdata.usgs.gov/nwis/help/?provisional for more information.",
                    "title": "disclaimer",
                },
                {"value": "caas01", "title": "server"},
            ],
        },
        "timeSeries": [
            {
                "sourceInfo": {
                    "siteName": "Traverse Creek near Kendall, PA",
                    "siteCode": [
                        {"value": "03107698", "network": "NWIS", "agencyCode": "USGS"}
                    ],
                    "timeZoneInfo": {
                        "defaultTimeZone": {
                            "zoneOffset": "-05:00",
                            "zoneAbbreviation": "EST",
                        },
                        "daylightSavingsTimeZone": {
                            "zoneOffset": "-04:00",
                            "zoneAbbreviation": "EDT",
                        },
                        "siteUsesDaylightSavingsTime": True,
                    },
                    "geoLocation": {
                        "geogLocation": {
                            "srs": "EPSG:4326",
                            "latitude": 40.5261111,
                            "longitude": -80.45138889,
                        },
                        "localSiteXY": [],
                    },
                    "note": [],
                    "siteType": [],
                    "siteProperty": [
                        {"value": "ST", "name": "siteTypeCd"},
                        {"value": "05030101", "name": "hucCd"},
                        {"value": "42", "name": "stateCd"},
                        {"value": "42007", "name": "countyCd"},
                    ],
                },
                "variable": {
                    "variableCode": [
                        {
                            "value": "00060",
                            "network": "NWIS",
                            "vocabulary": "NWIS:UnitValues",
                            "variableID": 45807197,
                            "default": True,
                        }
                    ],
                    "variableName": "Streamflow, ft&#179;/s",
                    "variableDescription": "Discharge, cubic feet per second",
                    "valueType": "Derived Value",
                    "unit": {"unitCode": "ft3/s"},
                    "options": {
                        "option": [
                            {
                                "value": "Mean",
                                "name": "Statistic",
                                "optionCode": "00003",
                            }
                        ]
                    },
                    "note": [],
                    "noDataValue": -999999.0,
                    "variableProperty": [],
                    "oid": "45807197",
                },
                "values": [
                    {
                        "value": [
                            {
                                "value": "6.82",
                                "qualifiers": ["A"],
                                "dateTime": "2017-04-23T00:00:00.000",
                            },
                            {
                                "value": "6.04",
                                "qualifiers": ["A"],
                                "dateTime": "2017-04-24T00:00:00.000",
                            },
                            {
                                "value": "5.62",
                                "qualifiers": ["A"],
                                "dateTime": "2017-04-25T00:00:00.000",
                            },
                            {
                                "value": "-999999",
                                "qualifiers": ["P", "Dis"],
                                "dateTime": "2017-04-26T00:00:00.000",
                            },
                            {
                                "value": "-999999",
                                "qualifiers": ["P", "Dis"],
                                "dateTime": "2017-04-26T00:00:00.000",
                            },
                            {
                                "value": "01234",
                                "qualifiers": ["P", "e"],
                                "dateTime": "2017-04-27T00:00:00.000",
                            },
                            {
                                "value": "-999999",
                                "qualifiers": ["P", "Dis"],
                                "dateTime": "2017-04-28T00:00:00.000",
                            },
                            {
                                "value": "-999999",
                                "qualifiers": ["P", "Dis"],
                                "dateTime": "2017-04-29T00:00:00.000",
                            },
                        ],
                        "qualifier": [
                            {
                                "qualifierCode": "Dis",
                                "qualifierDescription": "Record has been discontinued at the measurement site.",
                                "qualifierID": 0,
                                "network": "NWIS",
                                "vocabulary": "uv_rmk_cd",
                            },
                            {
                                "qualifierCode": "A",
                                "qualifierDescription": "Approved for publication -- Processing and review completed.",
                                "qualifierID": 1,
                                "network": "NWIS",
                                "vocabulary": "uv_rmk_cd",
                            },
                            {
                                "qualifierCode": "P",
                                "qualifierDescription": "Provisional data subject to revision.",
                                "qualifierID": 2,
                                "network": "NWIS",
                                "vocabulary": "uv_rmk_cd",
                            },
                            {
                                "qualifierCode": "e",
                                "qualifierDescription": "Estimated value.",
                                "qualifierID": 3,
                                "network": "NWIS",
                                "vocabulary": "uv_rmk_cd",
                            },
                        ],
                        "qualityControlLevel": [],
                        "method": [{"methodDescription": "", "methodID": 120332}],
                        "source": [],
                        "offset": [],
                        "sample": [],
                        "censorCode": [],
                    }
                ],
                "name": "USGS:03107698:00060:00003",
            }
        ],
    },
    "nil": False,
    "globalScope": True,
    "typeSubstituted": False,
}



# -*- coding: utf-8 -*-
"""
test_charts.py

Tests for the charts.py module.
"""
from __future__ import absolute_import, print_function, division, unicode_literals
import unittest
import matplotlib
import pandas as pd
print("Pandas version ", pd.__version__)

from hydrofunctions import charts
import hydrofunctions as hf
from .fixtures import JSON15min2day as test_json

dummy = {"col1": [1, 2, 3, 38, 23, 1, 19], "col2": [3, 4, 45, 23, 2, 4, 76]}


class TestFlowDuration(unittest.TestCase):
    def test_charts_flowduration_exists(self):
        expected = pd.DataFrame(data=dummy)
        actual_fig, actual_ax = charts.flow_duration(expected)
        self.assertIsInstance(actual_fig, matplotlib.figure.Figure)
        self.assertIsInstance(actual_ax, matplotlib.axes.Axes)

    def test_charts_flowduration_defaults(self):
        expected = pd.DataFrame(data=dummy)

        actual_fig, actual_ax = charts.flow_duration(expected)

        actual_xscale = actual_ax.xaxis.get_scale()
        actual_yscale = actual_ax.yaxis.get_scale()
        actual_ylabel = actual_ax.yaxis.get_label_text()
        actual_marker = actual_ax.get_lines()[0].get_marker()
        actual_legend = actual_ax.get_legend()
        actual_legend_loc = actual_legend._loc
        actual_title = actual_ax.get_title()

        self.assertEqual(actual_xscale, "logit")
        self.assertEqual(actual_yscale, "log")
        self.assertEqual(actual_ylabel, "Stream Discharge (m³/s)")
        self.assertEqual(actual_marker, ".")
        self.assertTrue(actual_legend)
        self.assertEqual(actual_legend_loc, 0)  # '0' is internal code for 'best'.
        self.assertEqual(actual_title, "")

    def test_charts_flowduration_accepts_params(self):
        expected = pd.DataFrame(data=dummy)
        params = {
            "xscale": "linear",
            "yscale": "linear",
            "ylabel": "test value",
            "symbol": ",",
            "legend": False,
            "legend_loc": "center",
            "title": "Test Title",
        }

        actual_fig, actual_ax = charts.flow_duration(expected, **params)

        actual_xscale = actual_ax.xaxis.get_scale()
        actual_yscale = actual_ax.yaxis.get_scale()
        actual_ylabel = actual_ax.yaxis.get_label_text()
        actual_marker = actual_ax.get_lines()[0].get_marker()
        actual_legend = actual_ax.get_legend()
        # There is no legend in this test, so there is no legend property.
        # actual_legend_loc = actual_legend._loc
        actual_title = actual_ax.get_title()

        self.assertEqual(actual_xscale, "linear")
        self.assertEqual(actual_yscale, "linear")
        self.assertEqual(actual_ylabel, "test value")
        self.assertEqual(actual_marker, ",")
        self.assertIsNone(actual_legend)
        # There is no legend, so there is no legend location property.
        # self.assertEqual(actual_legend_loc, 10) # 'center' is equal to 10.
        self.assertEqual(actual_title, "Test Title")


class TestCyclePlot(unittest.TestCase):
    def test_charts_groupby_not_object_dtype(self):
        # For reasons I don't understand, I think Pandas 0.25.0 counts
        # DataFrameGroupBy as an object, and you can't use .quintile() on it.?
        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
        self.assertFalse(pd.api.types.is_object_dtype(expected_df))
        grouped = expected_df.groupby(expected_df.index.weekday)
        self.assertIsInstance(grouped, pd.core.groupby.generic.DataFrameGroupBy)
        self.assertFalse(pd.api.types.is_object_dtype(grouped))

    def test_charts_cycleplot_exists(self):
        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
        # Select first numeric column
        expected_df = expected_df.loc[
            :, expected_df.select_dtypes(include="number").columns[0]
        ]
        actual_fig, actual_ax = charts.cycleplot(expected_df)
        self.assertIsInstance(actual_fig, matplotlib.figure.Figure)
        self.assertIsInstance(actual_ax[0], matplotlib.axes.Axes)

    def test_charts_cycleplot_raises_for_non_series(self):
        with self.assertRaises(ValueError):
            actual_fig, actual_ax = charts.cycleplot("nonsense")

    def test_charts_cycleplot_parts(self):
        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
        actual_fig, actual_ax = charts.cycleplot(
            expected_df, legend_loc="center", title="test title"
        )

        actual_xscale = actual_ax[0].xaxis.get_scale()
        actual_yscale = actual_ax[0].yaxis.get_scale()
        actual_ylabel = actual_ax[0].yaxis.get_label_text()
        actual_legend = actual_ax[0].get_legend()
        actual_legend_loc = actual_legend._loc
        actual_title = (
            actual_fig._suptitle.get_text()
        )  # unofficial title accessor! A little wonky.

        self.assertEqual(actual_xscale, "linear")
        self.assertEqual(actual_yscale, "linear")
        self.assertEqual(actual_ylabel, "Discharge (ft³/s)")
        self.assertTrue(actual_legend)
        self.assertEqual(
            actual_legend_loc, 10
        )  # '10' is internal code for legend(loc = 'center')
        self.assertEqual(actual_title, "test title")

    def test_charts_cycleplot_compare_month(self):
        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
        actual_fig, actual_ax = charts.cycleplot(expected_df, compare="month")
        self.assertIsInstance(actual_fig, matplotlib.figure.Figure)
        self.assertIsInstance(actual_ax[0], matplotlib.axes.Axes)

    def test_charts_cycleplot_cycle_annualweek(self):
        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
        actual_fig, actual_ax = charts.cycleplot(expected_df, "annual-week")
        self.assertIsInstance(actual_fig, matplotlib.figure.Figure)
        self.assertIsInstance(actual_ax[0], matplotlib.axes.Axes)

    def test_charts_cycleplot_cycle_annual(self):
        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
        actual_fig, actual_ax = charts.cycleplot(expected_df, "annual")
        self.assertIsInstance(actual_fig, matplotlib.figure.Figure)
        self.assertIsInstance(actual_ax[0], matplotlib.axes.Axes)

    def test_charts_cycleplot_cycle_annualdate(self):
        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
        actual_fig, actual_ax = charts.cycleplot(expected_df, "annual-date")
        self.assertIsInstance(actual_fig, matplotlib.figure.Figure)
        self.assertIsInstance(actual_ax[0], matplotlib.axes.Axes)

    def test_charts_cycleplot_cycle_annualmonth(self):
        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
        actual_fig, actual_ax = charts.cycleplot(expected_df, "annual-month")
        self.assertIsInstance(actual_fig, matplotlib.figure.Figure)
        self.assertIsInstance(actual_ax[0], matplotlib.axes.Axes)

    def test_charts_cycleplot_cycle_weekly(self):
        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
        actual_fig, actual_ax = charts.cycleplot(expected_df, "weekly")
        self.assertIsInstance(actual_fig, matplotlib.figure.Figure)
        self.assertIsInstance(actual_ax[0], matplotlib.axes.Axes)

    #    def test_charts_cycleplot_cycle_diurnalsmallest(self):
    #        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
    #        actual_fig, actual_ax = charts.cycleplot(expected_df, 'diurnal-smallest')
    #        self.assertIsInstance(actual_fig, matplotlib.figure.Figure)
    #        self.assertIsInstance(actual_ax[0], matplotlib.axes.Axes)

    def test_charts_cycleplot_cycle_diurnalhour(self):
        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
        actual_fig, actual_ax = charts.cycleplot(expected_df, "diurnal-hour")
        self.assertIsInstance(actual_fig, matplotlib.figure.Figure)
        self.assertIsInstance(actual_ax[0], matplotlib.axes.Axes)

    def test_charts_cycleplot_cycle_nonsense_raises_ValueError(self):
        expected_df, expected_dict = hf.extract_nwis_df(test_json, interpolate=False)
        with self.assertRaises(ValueError):
            actual_fig, actual_ax = charts.cycleplot(expected_df, "nonsense")


if __name__ == "__main__":
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 12 21:55:04 2016

test_exceptions.py
"""
from __future__ import absolute_import, print_function, division, unicode_literals
import unittest
import warnings

from hydrofunctions import exceptions


def raiseHydroNoDataError():
    raise exceptions.HydroNoDataError("test_exceptions.py raiseHydroNoDataError() ")


class TestExceptions(unittest.TestCase):
    def test_exceptions_HydroNoDataError_can_be_raised(self):
        self.assertRaises(exceptions.HydroNoDataError, raiseHydroNoDataError)

    def test_exceptions_HydroNoDataError_can_be_caught(self):
        actual = False
        try:
            raiseHydroNoDataError()
        except exceptions.HydroNoDataError as err:
            actual = True
        self.assertTrue(
            actual, "The HydroNoDataError should have been caught, but wasn't."
        )


class TestWarnings(unittest.TestCase):
    @unittest.skip(
        "assertWarns errors on Linux. See https://bugs.python.org/issue29620"
    )
    def test_exceptions_HydroUserWarning_can_be_called(self):
        with self.assertWarns(exceptions.HydroUserWarning):
            warnings.warn("test warning message", exceptions.HydroUserWarning)


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""
hydrofunctions.waterwatch
~~~~~~~~~~~~~~~~~~~~~~~~~

This module is for working with the five USGS WaterWatch Data Services.
Description of data services https://waterwatch.usgs.gov/webservices/

Main page: https://waterwatch.usgs.gov

NOTICE (taken from waterwatch.usgs.gov): In January 2020, USGS WaterWatch began 
operating in maintenance-only mode. Existing tools, features, and web data services are 
being fully maintained as before, but new tools and enhancements will no longer be 
developed. Please click here for more information or contact USGS WaterWatch if you have
any questions.

The WaterWatch program provides five data services with REST APIs:

- Current Conditions Real-Time Streamflow Service
- Flood and High Flow Service
- Average Streamflow for 7, 14, and 28 Days Service
- Hourly Flow Change Service
- Flood Stage Service

Hydrofunctions allows you to access each of these services as either a dictionary
or a dataframe with the station ID as the key/index.
-----
"""
import requests
import pandas as pd

WATERWATCH_URL = "https://waterwatch.usgs.gov/webservices/"


def _get_flood_stages(site_number: str = None):
    """Retrieves flood stages for USGS stations.

    Args:
        param site_number: Optional USGS station number. If not provided flood stages for all stations are retrieved.

    Returns: Dictionary of station numbers and corresponding flood stages
    """
    # response formats: json | xml | csv
    params = {"format": "json"}
    if site_number:
        params["site"] = site_number
    res = requests.get(WATERWATCH_URL + "floodstage", params=params)
    if res.ok:
        stages = res.json()
        return {
            site["site_no"]: {k: v for k, v in site.items() if k != "site_no"}
            for site in stages["sites"]
        }


def filter_flood_stages(all_flood_stages, sites_numbers=None):
    """Filters flood states of specific station numbers"""
    stations_stages = {}
    for site_nb in sites_numbers:
        try:
            stations_stages[site_nb] = all_flood_stages[site_nb]
        except KeyError:
            stations_stages[site_nb] = None
    return stations_stages


def get_flood_stage(site=None, output_format="dict"):
    """Retrieves flood stages for a list of station numbers.

    This function retrieves a dictionary of flood stages for each site. The 'stage' of
    a river is the height of the river surface at a stream gauge, expressed as a height
    above an arbitrary datum. It is similar to water depth, except that datums are
    usually set so that the zero (0) to be well below the lowest elevation of the
    stream bed. This is done so that even if there is erosion over time, the stream bed
    and the river stage will never reach an elevation that is less than zero. Stage is
    usually expressed in feet in this dataset. You can retrieve the stage of the river
    using the parameter '00065', whereas the discharge of the river is '00060'.

    There are several kinds of flood stage reported in these data:

        * action stage: If the water gets above this level, it triggers an action by
            the National Weather Service.
        * flood stage: Water at this level begins to be a hazard to lives, property, or
            commerce. Not necessarily the same as bankfull stage.
        * moderate flood stage: structures and roads begin to be inundated.
        * major flood stage: requires significant evacuations of people or transfer of
            property to higher elevations.

        See https://waterwatch.usgs.gov/webservices/ for more information.

    Args:
        site (str or list of str):
            The USGS site ID number or a list of numbers.

        output_format: Optional output format. Returns dict if 'dict' else returns pd.DataFrame

    Returns: Dictionary or DataFrame of station numbers and their flood stages. If
        there is no flood stage for a station, `None` is returned.

    Example:
        >>> stations = ["07144100", "07144101"]
        >>> res = get_flood_stage(stations, output_format="dict")  # dictionary output
        >>> print(res)
        {'07144100': {'action_stage': '20', 'flood_stage': '22', 'moderate_flood_stage': '25', 'major_flood_stage': '26'},
         '07144101': None}
        >>> print(get_flood_stage(stations))
    """
    all_stages = _get_flood_stages()

    if site:
        if isinstance(site, str):
            site = [site]
        stations_stages = filter_flood_stages(all_stages, site)
    else:
        stations_stages = all_stages

    if output_format == "dict":
        return stations_stages
    else:
        return pd.DataFrame(stations_stages).T



# -*- coding: utf-8 -*-
"""
Created on Tue Sep  6 11:20:51 2016

@author: Marty

test_validate.py
tests for va;idate.py
"""
from __future__ import absolute_import, print_function, division, unicode_literals
import unittest
from hydrofunctions import validate


class TestValidate(unittest.TestCase):
    def test_validate_check_parameter_string_None_returns_None(self):
        self.assertIsNone(validate.check_parameter_string(None, "site"))

    def test_validate_check_parameter_string_accepts_str_returns_str(self):
        test1 = "any string"
        test2 = "any string"
        test3 = "any unicode string"
        test4 = "a,comma,delimited,list"
        self.assertEqual(
            validate.check_parameter_string(test1, "site"), test1, "test 1 failed"
        )
        self.assertEqual(
            validate.check_parameter_string(test2, "site"), test2, "test 2 failed"
        )
        self.assertEqual(
            validate.check_parameter_string(test3, "site"), test3, "test 3 failed"
        )
        self.assertEqual(
            validate.check_parameter_string(test4, "site"), test4, "test 4 failed"
        )

    def test_validate_check_parameter_string_accepts_list_of_str(self):
        test1 = ["any string"]
        test2 = ["any string", "any string"]
        test3 = ["1", "2", "3", "4", "5", "6", "7", "8", "9"]
        # Why assert True?
        self.assertTrue(validate.check_parameter_string(test1, "site"), "test 1 failed")
        self.assertTrue(validate.check_parameter_string(test2, "site"), "test 2 failed")
        self.assertTrue(validate.check_parameter_string(test3, "site"), "test 3 failed")

    def test_validate_check_parameter_string_accepts_site_county_parameterCd(self):
        text = "text"
        self.assertTrue(validate.check_parameter_string(text, "site"), "test 1 failed")
        self.assertTrue(validate.check_parameter_string(text, "county"), "test 2 failed")
        self.assertTrue(
            validate.check_parameter_string(text, "parameterCd"), "test 3 failed"
        )

    def test_validate_check_parameter_string_raises_TypeError(self):
        not_string = 5
        self.assertRaises(TypeError, validate.check_parameter_string, not_string, "site")

    def test_validate_check_parameter_string_rejects_bad_list(self):
        test1 = []
        test2 = [""]
        test3 = [""]
        test4 = [4]
        test5 = ["good", None]
        test6 = ["good", 9]
        test7 = ["good", ""]
        self.assertRaises(TypeError, validate.check_parameter_string, test1)
        self.assertRaises(TypeError, validate.check_parameter_string, test2)
        self.assertRaises(TypeError, validate.check_parameter_string, test3)
        self.assertRaises(TypeError, validate.check_parameter_string, test4)
        self.assertRaises(TypeError, validate.check_parameter_string, test5)
        self.assertRaises(TypeError, validate.check_parameter_string, test6)
        self.assertRaises(TypeError, validate.check_parameter_string, test7)

    def test_validate_check_NWIS_bBox_accepts_list_tuple_string(self):
        b0 = (-105.43, 39.655, -104.0, 39.863)
        b1 = [-105.43, 39.655, -104.0, 39.863]
        bt = "-105.43,39.655,-104.0,39.863"
        self.assertEqual(validate.check_NWIS_bBox(b0), bt)
        self.assertEqual(validate.check_NWIS_bBox(b1), bt)
        self.assertEqual(validate.check_NWIS_bBox(bt), bt)

    def test_validate_bBox_None_returns_None(self):
        self.assertIsNone(validate.check_NWIS_bBox(None))

    def test_validate_bBox_accepts_list_of_four_latlongs(self):
        goodList = [-80, 20, -70, 30]
        actual = validate.check_NWIS_bBox(goodList)
        expected = "-80,20,-70,30"
        self.assertEqual(actual, expected)

    def test_validate_bBox_raises_TypeError_bad_input(self):
        bad1 = []
        bad2 = [1, 2, 3]  # We only test for insufficient args right now
        self.assertRaises(TypeError, validate.check_NWIS_bBox, bad1)
        self.assertRaises(TypeError, validate.check_NWIS_bBox, bad2)

    def test_validate_check_NWIS_service_accepts_iv_and_dv(self):
        self.assertTrue(validate.check_NWIS_service("iv"))
        self.assertEqual(validate.check_NWIS_service("iv"), "iv")
        self.assertTrue(validate.check_NWIS_service("dv"))
        self.assertEqual(validate.check_NWIS_service("dv"), "dv")

    def test_validate_check_NWIS_service_raises_TypeError(self):
        bad_input = "v"
        self.assertRaises(TypeError, validate.check_NWIS_service, bad_input)

    def test_validate_check_datestr_raises_TypeError(self):
        not_string = 5
        self.assertRaises(TypeError, validate.check_datestr, not_string)

    def test_validate_check_datestr_returns_true_for_good_date(self):
        # re pattern = r"[1-2]\d\d\d-[0-1]\d-[0-3]\d"
        actual = validate.check_datestr("2002-03-03")
        actual = validate.check_datestr("2002-03-03")
        self.assertTrue(actual)

    def test_validate_check_datestr_rejects_bad_date(self):
        # re pattern = r"[1-2]\d\d\d-[0-1]\d-[0-3]\d"
        bad1 = "3002-03-03"
        bad2 = "0000-03-03"
        bad3 = "1111-21-03"
        bad4 = "1111-03-43"
        self.assertRaises(TypeError, validate.check_datestr, bad1)
        self.assertRaises(TypeError, validate.check_datestr, bad2)
        self.assertRaises(TypeError, validate.check_datestr, bad3)
        self.assertRaises(TypeError, validate.check_datestr, bad4)

    def test_validate_check_period_accepts_good_period(self):
        # pattern = r^P\d{1,3}D$
        test1 = "P1D"
        test2 = "P12D"
        test3 = "P999D"  # Only accepts up to 999 days.
        self.assertEqual(validate.check_period(test1), test1)
        self.assertEqual(validate.check_period(test2), test2)
        self.assertEqual(validate.check_period(test3), test3)

    def test_validate_check_period_rejects_bad_period(self):
        # pattern = r^P\d{1,3}D$
        bad1 = "p34d"
        bad2 = "bad"
        bad3 = "P1234D"  # Too long
        bad4 = "P12"
        self.assertRaises(TypeError, validate.check_datestr, bad1)
        self.assertRaises(TypeError, validate.check_datestr, bad2)
        self.assertRaises(TypeError, validate.check_datestr, bad3)
        self.assertRaises(TypeError, validate.check_datestr, bad4)


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""
hydrofunctions.helpers
~~~~~~~~~~~~~~~~~~~~~~

This module holds functions designed to help out the user in an IPython
session.

-----
"""
from __future__ import absolute_import, print_function, division, unicode_literals
from IPython.core.display import HTML


def draw_map(width=700, height=400, url="http://hydrocloud.org"):
    """Draws a map of stream gages in a Jupyter Notebook.

    This function will draw an interactive map of stream gages from
    hydrocloud.org into an iframe and display it in a Jupyter Notebook.
    Each dot represents a stream gage. Click on the dot to learn its name,
    which you can use to request data.

    Args:
        width (int): The width of the map iframe.
        height (int): The height of the map iframe.
        url (str): The URL to put inside of the IFrame. Defaults to\
        https://hydrocloud.org

    Returns:
        HTML display object.

    **Example:**

        >>> import hydrofunctions as hf
        >>> hf.draw_map()

        A map appears.

        >>> hf.draw_map(width=900, height=600)

        Draws a larger map.
"""
    #    TODO:
    #        use ipywidgets to allow users to click on the map, and this will
    #        return a value that can be used in another IPython cell. This
    #        feature would allow the map to act as an interactive site selection
    #        tool.

    output = HTML(
        '<p>Use <a href="http://hydrocloud.org" target="_blank">'
        "HydroCloud.org</a> to find a stream gauge. "
        "Click on the dots to learn more about a site.</p>"
        "<iframe src=http://hydrocloud.org/ width={} height={}>"
        "</iframe>".format(width, height)
    )

    return output


def count_number_of_truthy(my_list):
    return sum(bool(item) for item in my_list)



# -*- coding: utf-8 -*-

"""
Hydrofunctions
~~~~~~~~~~~~~~

Hydrofunctions is a suite of convenience functions to help you explore
hydrology data interactively.

Basic Usage::

    >>> import hydrofunctions as hf

    >>> site = '01570500'
    >>> harrisburg = hf.NWIS(site, 'iv', period='P10D')
    Requested data from https://waterservices.usgs.gov/nwis/iv/?format=json%2C1.1&sites=01570500&period=P10D

    >>> harrisburg.ok
    True

Examine the dataset::

    >>> harrisburg
    USGS:01570500: Susquehanna River at Harrisburg, PA
        00045: <30 * Minutes> Precipitation, total, inches
        00060: <30 * Minutes> Discharge, cubic feet per second
        00065: <30 * Minutes> Gage height, feet
    Start: 2019-04-06 00:30:00+00:00
    End:   2019-04-15 23:00:00+00:00

The listing reports each of the parameters collected at the site that was
requested, how frequently the data are collected, and the name of the parameter
written out with units. The start and end of the dataset are given in Universal
Time (UTC).

View the first five rows of a dataframe that only contains the discharge data::

    >>> harrisburg.df('discharge').head()
                               USGS:01570500:00060:00000
    datetimeUTC
    2019-04-06 00:30:00+00:00                    44200.0
    2019-04-06 01:00:00+00:00                    44000.0
    2019-04-06 01:30:00+00:00                    44000.0
    2019-04-06 02:00:00+00:00                    43700.0
    2019-04-06 02:30:00+00:00                    43700.0

Because the .df() method returns a dataframe, you have access to all of the
methods associated with Pandas, including .plot(), .describe(), and .info() !

Learn more about hydrofunctions and the NWIS object with help()::

    >>> help(hf)
    >>> help(hf.NWIS)

Read more about Hydrofunctions here: https://hydrofunctions.readthedocs.io/

"""
from __future__ import absolute_import, print_function

__title__ = "hydrofunctions"
__version__ = "0.2.2"
__author__ = "Martin Roberge"
__email__ = "mroberge@towson.edu"
__license__ = "MIT"
__copyright__ = "Copyright 2016 Martin Roberge and contributors"


from .exceptions import (
    HydroNoDataError,
    HydroUserWarning,
)
from .hydrofunctions import (
    get_nwis,
    select_data,
    extract_nwis_df,
    nwis_custom_status_codes,
    get_nwis_property,
    calc_freq,
    read_parquet,
    save_parquet,
    read_json_gzip,
    save_json_gzip,
)
from .logging import _start_logging
from .station import Station, NWIS
from .validate import (
    check_parameter_string,
    check_datestr,
    check_NWIS_service,
    check_NWIS_bBox,
    check_period,
)
from .helpers import (
    draw_map,
    count_number_of_truthy,
)
from .charts import (
    flow_duration,
    cycleplot,
)
from .usgs_rdb import (
    get_usgs_RDB_service,
    read_rdb,
    site_file,
    data_catalog,
    field_meas,
    rating_curve,
    peaks,
    stats,
    hydroRDB,
)
from .waterwatch import (
    filter_flood_stages,
    get_flood_stage,
)



# -*- coding: utf-8 -*-
"""
fixtures_usgs_rdb.py

This file contains the following test fixtures:
    - field_fixture: a sample rdb file containing field measurements.
        - Header contains four 'messages'
        - dtypes contains one date column
        - 32 columns, 9 data rows (shortened from original)

    - rating_fixture: a sample rdb file containing an 'expanded shift-adjusted'
        rating curve.
        - Header has extra // marks to start every line. ???
        - Header has many different parameters embedded.
        - Two warning messages are in the header
        - 4 columns, 9 data rows (shortened from original)

    - peaks_fixture: a sample rdb file containing annual peak discharge data.
        - Header contains many 'messages', including two tables.
        - dtypes contain two date columns.
        13 columns, 18 data rows.

    - parsing_error_fixture: a peaks rdb file with a bad date of 1881-00-00.
        13 columns
        5 data rows, but 1 has a bad date.
"""


# shortened response from this URL: https://waterdata.usgs.gov/pa/nwis/measurements?site_no=01541200&agency_cd=USGS&format=rdb_expanded
field_fixture = """#
# U.S. Geological Survey, National Water Information System
# Surface water measurements
#
# Retrieved: 2019-03-04 10:15:59 EST     (sdww01)
#
# Further descriptions of the columns and codes used can be found at:
# https://help.waterdata.usgs.gov/output-formats#streamflow_measurement_data
#
# Data for the following 1 site(s) are contained in this file
#  USGS 01541200 WB Susquehanna River near Curwensville, PA
# -----------------------------------------------------------------------------------
#
#
agency_cd	site_no	measurement_nu	measurement_dt	tz_cd	q_meas_used_fg	party_nm	site_visit_coll_agency_cd	gage_height_va	discharge_va	measured_rating_diff	gage_va_change	gage_va_time	control_type_cd	discharge_cd	chan_nu	chan_name	meas_type	streamflow_method	velocity_method	chan_discharge	chan_width	chan_area	chan_velocity	chan_stability	chan_material	chan_evenness	long_vel_desc	horz_vel_desc	vert_vel_desc	chan_loc_cd	chan_loc_dist
5s	15s	6s	19d	12s	1s	12s	5s	12s	12s	12s	7s	6s	21s	15s	11n	64s	4s	5s	5s	14s	14s	14s	14s	4s	4s	4s	12s	9s	12s	7s	14s
USGS	01541200	1	1955-09-29		Yes	EVA	USGS	3.66	299	Good	-0.02	0.80	Clear	MEAS	1	Imported Channel 1	UNSP	other		299				UNSP	UNSP	UNSP	unkn	UNSP	UNSP	UNSP
USGS	01541200	2	1955-10-10		Yes	SNA	USGS	3.46	202	Good	0.00	1.00	Clear	MEAS	1	Imported Channel 1	UNSP	other		202				UNSP	UNSP	UNSP	unkn	UNSP	UNSP	UNSP
USGS	01541200	3	1955-11-17		Yes	GAR	USGS	6.77	3490	Good	-0.17	1.30	Clear	MEAS	1	Imported Channel 1	UNSP	other		3490				UNSP	UNSP	UNSP	unkn	UNSP	UNSP	UNSP
USGS	01541200	4	1955-12-18		Yes	YEH	USGS	3.64	193	Unspecified	0.00	0.80	Clear	MEAS	1	Imported Channel 1	UNSP	other		193				UNSP	UNSP	UNSP	unkn	UNSP	UNSP	UNSP
USGS	01541200	5	1956-01-11		Yes	YEH	USGS	3.52	152	Unspecified	0.00	1.00	Clear	MEAS	1	Imported Channel 1	UNSP	other		152				UNSP	UNSP	UNSP	unkn	UNSP	UNSP	UNSP
USGS	01541200	6	1956-02-15		Yes	GAR	USGS	4.87	1280	Good	0.10	1.20	Clear	MEAS	1	Imported Channel 1	UNSP	other		1280				UNSP	UNSP	UNSP	unkn	UNSP	UNSP	UNSP
USGS	01541200	7	1956-03-08		Yes	EVA	USGS	8.94	6750	Good	0.62	1.40	Clear	MEAS	1	Imported Channel 1	CRAN	unspe	unkno	6750	218	1610	4.19	UNSP	UNSP	UNSP	unkn	UNSP	UNSP	UNSP
USGS	01541200	8	1956-03-27		Yes	YEH	USGS	5.05	1410	Good	0.02	1.70	Clear	MEAS	1	Imported Channel 1	UNSP	other		1410				UNSP	UNSP	UNSP	unkn	UNSP	UNSP	UNSP
USGS	01541200	9	1956-04-23		Yes	WHY	USGS	5.08	1380	Good	0.01	1.20	Clear	MEAS	1	Imported Channel 1	UNSP	other		1380				UNSP	UNSP	UNSP	unkn	UNSP	UNSP	UNSP
"""

# shortened response from this URL: https://waterdata.usgs.gov/nwisweb/data/ratings/exsa/USGS.01541303.exsa.rdb
rating_fixture = """# //UNITED STATES GEOLOGICAL SURVEY       http://water.usgs.gov/
# //NATIONAL WATER INFORMATION SYSTEM     http://water.usgs.gov/data.html
# //DATA ARE PROVISIONAL AND SUBJECT TO CHANGE UNTIL PUBLISHED BY USGS
# //RETRIEVED: 2018-09-06 10:25:03
# //WARNING
# //WARNING The stage-discharge rating provided in this file should be
# //WARNING considered provisional and subject to change. Stage-discharge
# //WARNING ratings change over time as the channel features that control
# //WARNING the relation between stage and discharge vary. Users are
# //WARNING cautioned to consider carefully the applicability of this
# //WARNING rating before using it for decisions that concern personal or
# //WARNING public safety or operational consequences.
# //WARNING
# //WARNING This rating does not include any shifts that may have been
# //WARNING used along with this base rating in converting stage to
# //WARNING discharge at this site. Stage data processed with the rating
# //WARNING thus may not match that displayed or published by the USGS.
# //WARNING
# //FILE TYPE="NWIS RATING"
# //DATABASE NUMBER=01 DESCRIPTION=" Standard data base for this site."
# //STATION AGENCY="USGS " NUMBER="01541303       " TIME_ZONE="EST" DST_FLAG=Y
# //STATION NAME="West Branch Susquehanna River at Hyde, PA"
# //LABEL="Discharge (ft^3/s)"
# //PARAMETER CODE="00060"
# //RATING SHIFTED="20180906102503 EST"
# //RATING ID="11.0" TYPE="STGQ" NAME="stage-discharge" AGING=????
# //RATING REMARKS=""
# //RATING EXPANSION="logarithmic"
# //RATING OFFSET1=2.000000E+00
# //RATING_INDEP ROUNDING="????" PARAMETER="Gage height (ft)"
# //RATING_DEP ROUNDING="????" PARAMETER="Discharge (ft^3/s)"
# //RATING_DATETIME BEGIN=20160204024500 BZONE=-05:00 END=20170118000000 EZONE=-05:00 AGING=1
# //RATING_DATETIME BEGIN=20170118000000 BZONE=-05:00 END=-------------- EZONE=--- AGING=1
# //SHIFT_NEXT BEGIN="--------------" BZONE="---" END="--------------" EZONE="---"
# //SHIFT_NEXT STAGE1="---" SHIFT1="---" STAGE2="---" SHIFT2="---" STAGE3="---" SHIFT3="---"
# //SHIFT_NEXT COMMENT=" "
INDEP	SHIFT	DEP	STOR
16N	16N	16N	1S
2.50	0.00	27.90	*
2.51	0.00	29.25
2.52	0.00	30.65
2.53	0.00	32.08
2.54	0.00	33.54
2.55	0.00	35.05
2.56	0.00	36.59
2.57	0.00	38.18
2.58	0.00	39.80
"""


# A shortened version of this request: https://nwis.waterdata.usgs.gov/nwis/peak?site_no=01542500&agency_cd=USGS&format=rdb
peaks_fixture = """#
# U.S. Geological Survey
# National Water Information System
# Retrieved: 2019-10-28 22:35:01 EDT
#
# ---------------------------------- WARNING ----------------------------------------
# Some of the data that you have obtained from this U.S. Geological Survey database
# may not have received Director's approval. Any such data values are qualified
# as provisional and are subject to revision. Provisional data are released on the
# condition that neither the USGS nor the United States Government may be held liable
# for any damages resulting from its use.
#
# More data may be available offline.
# For more information on these data,  contact  USGS Water Data Inquiries.
# This file contains the annual peak streamflow data.
#
# This information includes the following fields:
#
#  agency_cd     Agency Code
#  site_no       USGS station number
#  peak_dt       Date of peak streamflow (format YYYY-MM-DD)
#  peak_tm       Time of peak streamflow (24 hour format, 00:00 - 23:59)
#  peak_va       Annual peak streamflow value in cfs
#  peak_cd       Peak Discharge-Qualification codes (see explanation below)
#  gage_ht       Gage height for the associated peak streamflow in feet
#  gage_ht_cd    Gage height qualification codes
#  year_last_pk  Peak streamflow reported is the highest since this year
#  ag_dt         Date of maximum gage-height for water year (if not concurrent with peak)
#  ag_tm         Time of maximum gage-height for water year (if not concurrent with peak
#  ag_gage_ht    maximum Gage height for water year in feet (if not concurrent with peak
#  ag_gage_ht_cd maximum Gage height code
#
# Sites in this file include:
#  USGS 01542500 WB Susquehanna River at Karthaus, PA
#
# Peak Streamflow-Qualification Codes(peak_cd):
#   1 ... Discharge is a Maximum Daily Average
#   2 ... Discharge is an Estimate
#   3 ... Discharge affected by Dam Failure
#   4 ... Discharge less than indicated value,
#           which is Minimum Recordable Discharge at this site
#   5 ... Discharge affected to unknown degree by
#           Regulation or Diversion
#   6 ... Discharge affected by Regulation or Diversion
#   7 ... Discharge is an Historic Peak
#   8 ... Discharge actually greater than indicated value
#   9 ... Discharge due to Snowmelt, Hurricane,
#           Ice-Jam or Debris Dam breakup
#   A ... Year of occurrence is unknown or not exact
#   Bd ... Day of occurrence is unknown or not exact
#   Bm ... Month of occurrence is unknown or not exact
#   C ... All or part of the record affected by Urbanization,
#            Mining, Agricultural changes, Channelization, or other
#   F ... Peak supplied by another agency
#   O ... Opportunistic value not from systematic data collection
#   R ... Revised
#
# Gage height qualification codes(gage_ht_cd,ag_gage_ht_cd):
#   1 ... Gage height affected by backwater
#   2 ... Gage height not the maximum for the year
#   3 ... Gage height at different site and(or) datum
#   4 ... Gage height below minimum recordable elevation
#   5 ... Gage height is an estimate
#   6 ... Gage datum changed during this year
#   7 ... Debris, mud, or hyper-concentrated flow
#   8 ... Gage height tidally affected
#   Bd ... Day of occurrence is unknown or not exact
#   Bm ... Month of occurrence is unknown or not exact
#   F ... Peak supplied by another agency
#   R ... Revised
#
#
agency_cd	site_no	peak_dt	peak_tm	peak_va	peak_cd	gage_ht	gage_ht_cd	year_last_pk	ag_dt	ag_tm	ag_gage_ht	ag_gage_ht_cd
5s	15s	10d	6s	8s	33s	8s	27s	4s	10d	6s	8s	27s
USGS	01542500	1936-03-18		135000	7	24.50
USGS	01542500	1940-04-01		50900		12.40	3
USGS	01542500	1941-04-06		19600		8.79	2		1941-03-05		8.95	1
USGS	01542500	1942-03-10		22600		9.38
USGS	01542500	1942-12-30		50200		13.82
USGS	01542500	1962-04-01		17000	6	8.35
USGS	01542500	1963-03-18		22700	6	9.63
USGS	01542500	1964-03-10		63500	6	15.98
USGS	01542500	1965-03-30		13600	6	7.54
USGS	01542500	1966-02-14		18900	6	8.88	2		1966-02-13		9.19	1
USGS	01542500	1967-03-16		17400	6	8.49
USGS	01542500	1968-05-25		11800	6	7.12	2		1968-01-31		7.12	1
USGS	01542500	1968-12-29		9500	6	6.35
USGS	01542500	1970-04-02		25800	6	10.29
USGS	01542500	1971-02-28		18400	6	8.79
USGS	01542500	2016-02-04		7880	6	6.05
USGS	01542500	2017-05-30		15700	6	8.43
USGS	01542500	2018-09-10	21:30	41000	6	13.22
"""

#The following is shortened from: https://nwis.waterdata.usgs.gov/nwis/peak?site_no=06813500&agency_cd=USGS&format=rdb
# You can reproduce this with: hf.peaks('06813500')
parsing_error_fixture = """#
# U.S. Geological Survey
# National Water Information System
# Retrieved: 2021-03-24 21:15:41 EDT
#
# ---------------------------------- WARNING ----------------------------------------
# Some of the data that you have obtained from this U.S. Geological Survey database
# may not have received Director's approval. Any such data values are qualified
# as provisional and are subject to revision. Provisional data are released on the
# condition that neither the USGS nor the United States Government may be held liable
# for any damages resulting from its use.
#
# More data may be available offline.
# For more information on these data,  contact  USGS Water Data Inquiries.
# This file contains the annual peak streamflow data.
#
# This information includes the following fields:
#
#  agency_cd     Agency Code
#  site_no       USGS station number
#  peak_dt       Date of peak streamflow (format YYYY-MM-DD)
#  peak_tm       Time of peak streamflow (24 hour format, 00:00 - 23:59)
#  peak_va       Annual peak streamflow value in cfs
#  peak_cd       Peak Discharge-Qualification codes (see explanation below)
#  gage_ht       Gage height for the associated peak streamflow in feet
#  gage_ht_cd    Gage height qualification codes
#  year_last_pk  Peak streamflow reported is the highest since this year
#  ag_dt         Date of maximum gage-height for water year (if not concurrent with peak)
#  ag_tm         Time of maximum gage-height for water year (if not concurrent with peak
#  ag_gage_ht    maximum Gage height for water year in feet (if not concurrent with peak
#  ag_gage_ht_cd maximum Gage height code
#
# Sites in this file include:
#  USGS 06813500 Missouri River at Rulo, NE
#
# Peak Streamflow-Qualification Codes(peak_cd):
#   1 ... Discharge is a Maximum Daily Average
#   2 ... Discharge is an Estimate
#   3 ... Discharge affected by Dam Failure
#   4 ... Discharge less than indicated value,
#           which is Minimum Recordable Discharge at this site
#   5 ... Discharge affected to unknown degree by
#           Regulation or Diversion
#   6 ... Discharge affected by Regulation or Diversion
#   7 ... Discharge is an Historic Peak
#   8 ... Discharge actually greater than indicated value
#   9 ... Discharge due to Snowmelt, Hurricane,
#           Ice-Jam or Debris Dam breakup
#   A ... Year of occurrence is unknown or not exact
#   Bd ... Day of occurrence is unknown or not exact
#   Bm ... Month of occurrence is unknown or not exact
#   C ... All or part of the record affected by Urbanization,
#            Mining, Agricultural changes, Channelization, or other
#   F ... Peak supplied by another agency
#   O ... Opportunistic value not from systematic data collection
#   R ... Revised
#
# Gage height qualification codes(gage_ht_cd,ag_gage_ht_cd):
#   1 ... Gage height affected by backwater
#   2 ... Gage height not the maximum for the year
#   3 ... Gage height at different site and(or) datum
#   4 ... Gage height below minimum recordable elevation
#   5 ... Gage height is an estimate
#   6 ... Gage datum changed during this year
#   7 ... Debris, mud, or hyper-concentrated flow
#   8 ... Gage height tidally affected
#   Bd ... Day of occurrence is unknown or not exact
#   Bm ... Month of occurrence is unknown or not exact
#   F ... Peak supplied by another agency
#   R ... Revised
#
#
agency_cd	site_no	peak_dt	peak_tm	peak_va	peak_cd	gage_ht	gage_ht_cd	year_last_pk	ag_dt	ag_tm	ag_gage_ht	ag_gage_ht_cd
5s	15s	10d	6s	8s	33s	8s	27s	4s	10d	6s	8s	27s
USGS	06813500	1881-00-00				22.90	Bm					
USGS	06813500	1950-04-29		185000					1950-04-30		21.60	
USGS	06813500	1951-06-03		175000					1951-05-02		20.90	
USGS	06813500	1952-04-22		358000		25.60						
USGS	06813500	1953-06-28		117000	6	17.47						
"""



# -*- coding: utf-8 -*-
"""
Created on Mon Sep  5 21:13:34 2016

@author: Marty
"""

from __future__ import absolute_import, print_function, division, unicode_literals
import unittest
from unittest import mock
import pandas as pd
from pandas.testing import assert_frame_equal
import numpy as np

from hydrofunctions import station, validate
from .fixtures import (
    fakeResponse,
    recent_only,
)


class MockNWIS(station.NWIS):
    """
    This subclass of NWIS is for testing all of the NWIS methods except
    __init__, which we will replace. All of the other methods get inherited
    verbatim, so we can test them using MockNWIS instead of NWIS.
    """

    def __init__(
        self,
        site=None,
        service=None,
        start_date=None,
        end_date=None,
        dataframe=None,
        meta=None,
        start=None,
        end=None,
    ):
        self.site = site
        self.service = service
        self.start_date = start_date
        self.end_date = end_date
        self._dataframe = dataframe
        self.meta = meta
        self.start = start
        self.end = end


class TestStation(unittest.TestCase):
    def test_station_is_obj(self):
        actual = station.Station()
        self.assertIsInstance(actual, station.Station)

    def test_station_site_defaults_to_None(self):
        actual = station.Station()
        self.assertIsNone(actual.site)

    def test_station_id_sets(self):
        expected = "01234567"
        actual = station.Station(expected)
        another = station.Station("23456789")
        self.assertEqual(actual.site, expected)
        self.assertEqual(another.site, "23456789")

    def test_station_dict_returns_dict(self):
        actual = station.Station("first")
        self.assertIsInstance(actual.station_dict, dict)

    def test_multiple_instances_only_one_list(self):
        first = station.Station("first")
        second = station.Station("second")
        self.assertEqual(first.station_dict, second.station_dict)

    def test_station_dict_keeps_keys(self):
        first = station.Station("first")
        second = station.Station("second")
        actual = first.station_dict
        self.assertIn("first", actual)
        self.assertIn("second", actual)
        self.assertEqual(
            len(actual),
            2,
            "The dict length is not equal to the number of instances",
        )

    def test_station_dict_returns_instance(self):
        first = station.Station("first")
        second = station.Station("second")
        expected = first
        # Look at the station_dict; does it contain a ref to 'first'?
        actual = second.station_dict["first"]
        self.assertEqual(actual, expected)

    def test_station_subclasses_maintain_same_station_dict(self):
        class Foo(station.Station):
            pass

        foo_inst = Foo("foo")
        station_inst = station.Station("station")
        self.assertIn("station", foo_inst.station_dict)
        self.assertIn("foo", station_inst.station_dict)
        actual = station_inst.station_dict["foo"]
        self.assertIsInstance(actual, Foo)


class TestNWISinit(unittest.TestCase):
    @mock.patch("hydrofunctions.hydrofunctions.get_nwis")
    @mock.patch("hydrofunctions.hydrofunctions.get_nwis_property")
    def test_NWIS_init_check_defaults(self, mock_get_nwis_property, mock_get_nwis):
        default_site = None
        default_service = "dv"
        default_start = None
        default_end = None
        default_parameterCd = "all"
        default_period = None
        default_stateCd = None
        default_countyCd = None
        default_bBox = None
        default_verbose = True

        mock_get_nwis_property.return_value = "expected"
        mock_get_nwis.return_value = fakeResponse()

        station.NWIS()

        mock_get_nwis.assert_called_once_with(
            default_site,
            default_service,
            default_start,
            default_end,
            parameterCd=default_parameterCd,
            period=default_period,
            stateCd=default_stateCd,
            countyCd=default_countyCd,
            bBox=default_bBox,
            verbose=default_verbose,
        )
        self.assertTrue(mock_get_nwis_property)

    @mock.patch("hydrofunctions.hydrofunctions.get_nwis")
    @mock.patch("hydrofunctions.hydrofunctions.get_nwis_property")
    def test_NWIS_init_calls_get_nwis_and_get_prop(
        self, mock_get_nwis_property, mock_get_nwis
    ):
        site = "expected site"
        service = "expected service"
        start = "expected start"
        end = "expected end"
        parameterCd = "expected pCode"
        mock_get_nwis_property.return_value = "expected"
        mock_get_nwis.return_value = fakeResponse()

        station.NWIS(site, service, start, end, parameterCd=parameterCd)
        mock_get_nwis.assert_called_once_with(
            site,
            service,
            start,
            end,
            parameterCd=parameterCd,
            period=None,
            stateCd=None,
            countyCd=None,
            bBox=None,
            verbose=True,
        )
        self.assertTrue(mock_get_nwis_property)

    @mock.patch("hydrofunctions.hydrofunctions.get_nwis")
    @mock.patch("hydrofunctions.hydrofunctions.get_nwis_property")
    @mock.patch("hydrofunctions.hydrofunctions.extract_nwis_df")
    def test_NWIS_init_sets_url_ok_json(
        self, mock_extract_nwis_df, mock_get_nwis_property, mock_get_nwis
    ):
        expected_url = "expected url"
        expected_ok = True
        expected_json = "expected json"

        mock_get_nwis.return_value = fakeResponse(
            code=200, url=expected_url, json=expected_json
        )
        mock_df = pd.DataFrame(
            np.random.randn(5, 1),
            columns=["A"],
            index=pd.date_range("20130101", periods=5, freq="T"),
        )
        mock_extract_nwis_df.return_value = (mock_df, "expected_dict")

        actual = station.NWIS()
        # self.assertEqual(actual.url, expected_url, "NWIS.__init__() did not set self.url properly.")
        self.assertEqual(
            actual.ok, expected_ok, "NWIS.__init__() did not set self.ok properly."
        )
        self.assertEqual(
            actual.json,
            expected_json,
            "NWIS.__init__() did not set self.json properly.",
        )

    @mock.patch("hydrofunctions.hydrofunctions.get_nwis")
    @mock.patch("hydrofunctions.hydrofunctions.get_nwis_property")
    @mock.patch("hydrofunctions.hydrofunctions.extract_nwis_df")
    def test_NWIS_init_calls_extract_nwis_df(
        self, mock_extract_nwis_df, mock_get_nwis_property, mock_get_nwis
    ):
        expected_json = "expected json"
        mock_get_nwis.return_value = fakeResponse(json=expected_json)
        mock_df = pd.DataFrame(
            np.random.randn(5, 1),
            columns=["A"],
            index=pd.date_range("20130101", periods=5, freq="T"),
        )
        mock_extract_nwis_df.return_value = (mock_df, "expected dict")
        actual = station.NWIS()
        mock_extract_nwis_df.assert_called_once_with(expected_json)

    @mock.patch("hydrofunctions.hydrofunctions.read_parquet")
    def test_NWIS_init_filename_calls_read_parquet(self, mock_read):
        expected_filename = "expected_filename.parquet"
        expected_meta = "expected meta"
        expected_df = pd.DataFrame(
            np.random.randn(5, 1),
            columns=["A"],
            index=pd.date_range("20130101", periods=5, freq="T"),
        )
        mock_start = "expected start"
        mock_end = "expected end"
        mock_read.return_value = (expected_df, expected_meta)
        actual = station.NWIS(file=expected_filename)
        mock_read.assert_called_once_with(expected_filename)
        assert_frame_equal(expected_df, actual._dataframe)
        self.assertEqual(
            expected_meta,
            actual.meta,
            "The metadata were not retrieved by NWIS.read().",
        )

    @mock.patch("hydrofunctions.hydrofunctions.read_parquet")
    @mock.patch("hydrofunctions.hydrofunctions.get_nwis")
    # @mock.patch("hydrofunctions.hydrofunctions.get_nwis_property")
    @mock.patch("hydrofunctions.hydrofunctions.extract_nwis_df")
    @mock.patch("hydrofunctions.hydrofunctions.save_parquet")
    def test_NWIS_init_filename_calls_read_parquet_then_get_nwis(
        self, mock_save, mock_extract_nwis_df, mock_get_nwis, mock_read
    ):
        # Mocks listed in order that they get called.

        # mock_read: pretend file doesn't exist, so return OSError
        #    file exists:
        #         mock_read.return_value = (expected_df, expected_meta)
        #    file doesn't exist, raise error:
        mock_read.side_effect = OSError()

        # mock_get_nwis
        expected_json = "expected json"
        mock_get_nwis.return_value = fakeResponse(json=expected_json)

        # mock_get_nwis_property
        # never called

        # mock_extract_nwis_df
        mock_df = pd.DataFrame(
            np.random.randn(5, 1),
            columns=["A"],
            index=pd.date_range("20130101", periods=5, freq="T"),
        )
        mock_meta = "mock meta"
        mock_extract_nwis_df.return_value = (mock_df, mock_meta)

        # mock_save
        expected_filename = "expected_filename.parquet"
        mock_save.return_value = "expected self"

        # Create an NWIS with a filename, but the filename doesn't exist.
        # so an OSError is returned.
        # So now get_nwis is called, extract_nwis_df, save().
        actual = station.NWIS(file=expected_filename)
        mock_save.assert_called_once_with(expected_filename, mock_df, mock_meta)

    @mock.patch("hydrofunctions.hydrofunctions.get_nwis")
    def test_NWIS_init_request_most_recent_only(self, mock_get_nwis):
        expected_json = recent_only
        expected_url = (
            "https://waterservices.usgs.gov/nwis/dv/?format=json%2C1.1&sites=01541200"
        )
        mock_get_nwis.return_value = fakeResponse(json=expected_json, url=expected_url)
        actual = station.NWIS("01541200")
        self.assertEqual(
            actual.df().shape,
            (2, 4),
            "The dataframe should only have two rows and four columns.",
        )


class TestNWISmethods(unittest.TestCase):
    """
    Tests for NWIS methods
    The following section is for testing all of the NWIS methods
    EXCEPT the NWIS.__init__() method.

    Creating an NWIS instance will always run the __init__ method, which we
    usually don't want to do. It calls a bunch of functions that we test
    elsewhere and it causes a bunch of side effects that we don't want. Yes,
    you can mock all of the functions that __init__ calls, but even then there
    can be unwanted side effects not to mention it can be tedious to mock so
    many different things.

    To test any method other than __init__, we will use the following strategy:
        - create a sub-class of NWIS called MockNWIS.
        - MockNWIS has a different __init__ method that allows you to pass
            in a dataframe and any other initial parameter
        - all other methods gets inherited from NWIS, so we can test them."""

    def test_NWIS_df_returns_all_columns(self):
        expected_cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]
        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=expected_cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df()
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols, expected_cols, "NWIS.df() should return all of the columns."
        )

    def test_NWIS_df_all_returns_all_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]
        expected_cols = cols
        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("all")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('all') should return all of the columns.",
        )

    def test_NWIS_df_data_returns_data_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]
        expected_cols = [
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("data")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('data') should return all of the data columns.",
        )

    def test_NWIS_df_discharge_returns_discharge_data_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = ["USGS:01541200:00060:00000", "USGS:01541303:00060:00000"]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("discharge")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('discharge') should return all of the discharge data columns.",
        )

    def test_NWIS_df_q_returns_discharge_data_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = ["USGS:01541200:00060:00000", "USGS:01541303:00060:00000"]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("q")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('q') should return all of the discharge data columns.",
        )

    def test_NWIS_df_stage_returns_stage_data_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = ["USGS:01541200:00065:00000", "USGS:01541303:00065:00000"]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("stage")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('stage') should return all of the stage data columns.",
        )

    def test_NWIS_df_q_raises_ValueError_when_no_q(self):

        cols = [
            "USGS:01541200:00011:00000_qualifiers",
            "USGS:01541200:00011:00000",
            "USGS:01541200:00022:00000_qualifiers",
            "USGS:01541200:00022:00000",
            "USGS:01541303:00011:00000_qualifiers",
            "USGS:01541303:00011:00000",
            "USGS:01541303:00022:00000_qualifiers",
            "USGS:01541303:00022:00000",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        shortcut_test_parameters = ["q", "discharge", "stage"]
        for shortcut in shortcut_test_parameters:
            with self.subTest(shortcut=shortcut):
                # Pytest does not run unittest sub-tests properly; only the first case fails.
                with self.assertRaises(
                    ValueError,
                    msg=f"Requesting '{shortcut}' when this parameter isn't collected should raise a ValueError.",
                ):
                    actual_df = test_nwis.df(shortcut)

    def test_NWIS_df_flags_returns_qualifiers_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00065:00000_qualifiers",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("flags")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('flags') should return all of the qualifier columns.",
        )

    def test_NWIS_df_flags_q_returns_discharge_qualifiers_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541303:00060:00000_qualifiers",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("flags", "q")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('flags', 'q') should return all of the discharge qualifier columns.",
        )

    def test_NWIS_df_stage_flags_returns_stage_qualifiers_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = [
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541303:00065:00000_qualifiers",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("stage", "flags")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('stage', 'flags') should return all of the stage qualifier columns.",
        )

    def test_NWIS_df_crazy_input_raises_ValueError(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)

        with self.assertRaises(
            ValueError,
            msg="unmatched args such as NWIS.df('crazy', 'input') should cause NWIS.df() to raise a ValueError.",
        ):
            actual_df = test_nwis.df("discharge", "crazy", "input")

    def test_NWIS_df_5digits_returns_param_data_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = ["USGS:01541200:00065:00000", "USGS:01541303:00065:00000"]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("00065")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('00065') should return all of the 00065 data columns.",
        )

    def test_NWIS_df_5digits_and_flags_returns_00065_qualifiers_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = [
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541303:00065:00000_qualifiers",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("00065", "flags")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('00065', 'flags') should return all of the 00065 _qualifiers columns.",
        )

    def test_NWIS_df_5digits_no_match_raises_ValueError(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)

        with self.assertRaises(
            ValueError,
            msg="A five-digit input that doesn't match a paramCode should raise a ValueError.",
        ):
            actual_df = test_nwis.df("00000")

    def test_NWIS_df_6digits_raises_ValueError(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)

        with self.assertRaises(
            ValueError,
            msg="A six-digit input like .df('123456') should raise a ValueError.",
        ):
            actual_df = test_nwis.df("123456")

    def test_NWIS_df_7digits_raises_ValueError(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)

        with self.assertRaises(
            ValueError,
            msg="A seven-digit input like .df('1234567') should raise a ValueError.",
        ):
            actual_df = test_nwis.df("1234567")

    def test_NWIS_df_8digits_returns_station_data_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = ["USGS:01541200:00060:00000", "USGS:01541200:00065:00000"]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("01541200")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('01541200') should return all of the data columns for station 01541200.",
        )

    def test_NWIS_df_8digits_and_flags_returns_station_qualifiers_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00065:00000_qualifiers",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("01541200", "flags")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('01541200', 'flags') should return all of the _qualifiers columns for station '01541200'.",
        )

    def test_NWIS_df_8digits_no_match_raises_ValueError(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)

        with self.assertRaises(
            ValueError,
            msg="An eight-digit input that doesn't match a station id should raise a ValueError.",
        ):
            actual_df = test_nwis.df("12345678")

    def test_NWIS_df_two_sites_returns_two_site_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = [
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("01541200", "01541303")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('01541200', '01541303') should return data columns for two sites.",
        )

    def test_NWIS_df_two_params_returns_two_param_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = [
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("00060", "00065")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('00060', '00065') should return data columns for two parameters.",
        )

    def test_NWIS_df_flag_and_data_returns_flags_and_data_columns(self):
        cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
            "USGS:01541200:00065:00000_qualifiers",
            "USGS:01541200:00065:00000",
            "USGS:01541303:00060:00000_qualifiers",
            "USGS:01541303:00060:00000",
            "USGS:01541303:00065:00000_qualifiers",
            "USGS:01541303:00065:00000",
        ]

        expected_cols = [
            "USGS:01541200:00060:00000_qualifiers",
            "USGS:01541200:00060:00000",
        ]

        data = [
            ["test", 5, "test", 5, "test", 5, "test", 5],
            ["test", 5, "test", 5, "test", 5, "test", 5],
        ]
        test_df = pd.DataFrame(data=data, columns=cols)
        test_nwis = MockNWIS(dataframe=test_df)
        actual_df = test_nwis.df("flags", "data", "discharge", "01541200")
        actual_cols = actual_df.columns.tolist()
        self.assertListEqual(
            actual_cols,
            expected_cols,
            "NWIS.df('flags', 'data', 'discharge', '01541200') should return data columns and flags for one site and one parameter.",
        )

    def test_NWIS_repr_returns_str(self):
        mock_meta = {
            "USGS:01541200": {
                "siteName": "WB Susquehanna River near Curwensville, PA",
                "siteLatLongSrs": {
                    "srs": "EPSG:4326",
                    "latitude": 40.9614471,
                    "longitude": -78.5191906,
                },
                "timeSeries": {
                    "00060": {
                        "variableFreq": "<Day>",
                        "variableUnit": "ft3/s",
                        "variableDescription": "Discharge, cubic feet per second",
                        "methodDescription": "",
                    }
                },
            }
        }
        mock_start = "expected start"
        mock_end = "expected end"
        test_nwis = MockNWIS(meta=mock_meta, start=mock_start, end=mock_end)
        expected_repr = "USGS:01541200: WB Susquehanna River near Curwensville, PA\n    00060: <Day>  Discharge, cubic feet per second \nStart: expected start\nEnd:   expected end"
        self.assertEqual(repr(test_nwis), expected_repr)

    def test_NWIS_repr_multisite_multi_param_returns_repr(self):
        self.maxDiff = None
        mock_meta = {
            "USGS:01585200": {
                "siteName": "WEST BRANCH HERRING RUN AT IDLEWYLDE, MD",
                "siteLatLongSrs": {
                    "srs": "EPSG:4326",
                    "latitude": 39.37363889,
                    "longitude": -76.5843333,
                },
                "timeSeries": {
                    "00060": {
                        "variableFreq": "<5 * Minutes>",
                        "variableUnit": "ft3/s",
                        "variableDescription": "Discharge, cubic feet per second",
                        "methodDescription": "",
                    },
                    "00065": {
                        "variableFreq": "<5 * Minutes>",
                        "variableUnit": "ft",
                        "variableDescription": "Gage height, feet",
                        "methodDescription": "",
                    },
                },
            },
            "USGS:01585219": {
                "siteName": "HERRING RUN AT SINCLAIR LANE AT BALTIMORE, MD",
                "siteLatLongSrs": {
                    "srs": "EPSG:4326",
                    "latitude": 39.31796389,
                    "longitude": -76.5551306,
                },
                "timeSeries": {
                    "00010": {
                        "variableFreq": "<5 * Minutes>",
                        "variableUnit": "deg C",
                        "variableDescription": "Temperature, water, degrees Celsius",
                        "methodDescription": "",
                    },
                    "00060": {
                        "variableFreq": "<5 * Minutes>",
                        "variableUnit": "ft3/s",
                        "variableDescription": "Discharge, cubic feet per second",
                        "methodDescription": "",
                    },
                    "00065": {
                        "variableFreq": "<5 * Minutes>",
                        "variableUnit": "ft",
                        "variableDescription": "Gage height, feet",
                        "methodDescription": "",
                    },
                    "00095": {
                        "variableFreq": "<5 * Minutes>",
                        "variableUnit": "uS/cm @25C",
                        "variableDescription": "Specific conductance, water, unfiltered, microsiemens per centimeter at 25 degrees Celsius",
                        "methodDescription": "",
                    },
                    "00400": {
                        "variableFreq": "<5 * Minutes>",
                        "variableUnit": "std units",
                        "variableDescription": "pH, water, unfiltered, field, standard units",
                        "methodDescription": "",
                    },
                },
            },
        }
        mock_start = "expected start"
        mock_end = "expected end"
        test_nwis = MockNWIS(meta=mock_meta, start=mock_start, end=mock_end)
        expected_repr = """USGS:01585200: WEST BRANCH HERRING RUN AT IDLEWYLDE, MD
    00060: <5 * Minutes>  Discharge, cubic feet per second 
    00065: <5 * Minutes>  Gage height, feet 
USGS:01585219: HERRING RUN AT SINCLAIR LANE AT BALTIMORE, MD
    00010: <5 * Minutes>  Temperature, water, degrees Celsius 
    00060: <5 * Minutes>  Discharge, cubic feet per second 
    00065: <5 * Minutes>  Gage height, feet 
    00095: <5 * Minutes>  Specific conductance, water, unfiltered, microsiemens per centimeter at 25 degrees Celsius 
    00400: <5 * Minutes>  pH, water, unfiltered, field, standard units 
Start: expected start
End:   expected end"""
        self.assertEqual(repr(test_nwis), expected_repr)

    @mock.patch("hydrofunctions.hydrofunctions.save_parquet")
    def test_NWIS_save_calls_save_parquet(self, mock_save):
        expected_filename = "expected_filename.parquet"
        expected_meta = "expected meta"
        expected_df = "expected df"
        mock_start = "expected start"
        mock_end = "expected end"
        test_nwis = MockNWIS(
            dataframe=expected_df, meta=expected_meta, start=mock_start, end=mock_end
        )
        test_nwis.save(expected_filename)
        mock_save.assert_called_once_with(expected_filename, expected_df, expected_meta)

    @mock.patch("hydrofunctions.hydrofunctions.save_json_gzip")
    def test_NWIS_save_calls_save_gz(self, mock_save):
        expected_filename = "expected_filename.gz"
        expected_json = "expected json"
        test_nwis = MockNWIS()
        test_nwis.json = expected_json
        test_nwis.save(expected_filename)
        mock_save.assert_called_once_with(expected_filename, expected_json)

    def test_NWIS_save_gz_raises_error_if_no_json(self):
        expected_filename = "expected_filename.gz"
        test_nwis = MockNWIS()
        with self.assertRaises(AttributeError):
            test_nwis.save(expected_filename)

    def test_NWIS_save_gz_raises_error_if_unrecognized_file_extension(self):
        expected_filename = "expected_filename.weird"
        test_nwis = MockNWIS()
        with self.assertRaises(OSError):
            test_nwis.save(expected_filename)

    @mock.patch("hydrofunctions.hydrofunctions.read_parquet")
    def test_NWIS_read_calls_read_parquet(self, mock_read):
        expected_filename = "expected_filename.parquet"
        expected_meta = "expected meta"
        expected_df = "expected df"
        mock_start = "expected start"
        mock_end = "expected end"
        mock_read.return_value = (expected_df, expected_meta)
        test_nwis = MockNWIS()
        test_nwis.read(expected_filename)
        mock_read.assert_called_once_with(expected_filename)
        self.assertEqual(
            expected_df,
            test_nwis._dataframe,
            "NWIS.read() did not retrieve the dataframe properly.",
        )
        self.assertEqual(
            expected_meta,
            test_nwis.meta,
            "The metadata were not retrieved by NWIS.read().",
        )

    @mock.patch("hydrofunctions.hydrofunctions.read_json_gzip")
    def test_NWIS_read_calls_read_json_gz(self, mock_read):
        expected_filename = "expected_filename.json.gz"
        expected_json = recent_only
        mock_read.return_value = expected_json
        expected_meta = {
            "USGS:01541200": {
                "siteName": "WB Susquehanna River near Curwensville, PA",
                "siteLatLongSrs": {
                    "srs": "EPSG:4326",
                    "latitude": 40.9614471,
                    "longitude": -78.5191906,
                },
                "timeSeries": {
                    "00010": {
                        "variableFreq": "<0 * Minutes>",
                        "variableUnit": "deg C",
                        "variableDescription": "Temperature, water, degrees Celsius",
                        "methodID": "118850",
                        "methodDescription": "",
                    },
                    "00060": {
                        "variableFreq": "<0 * Minutes>",
                        "variableUnit": "ft3/s",
                        "variableDescription": "Discharge, cubic feet per second",
                        "methodID": "118849",
                        "methodDescription": "",
                    },
                },
            }
        }
        test_nwis = MockNWIS()
        test_nwis.read(expected_filename)
        mock_read.assert_called_once_with(expected_filename)
        actual_meta = test_nwis.meta
        self.assertEqual(
            actual_meta,
            expected_meta,
            f"The read function did not parse the json.gz file correctly. It returned"
            f"type {type(actual_meta)} when it was expected to return type {type(expected_meta)}.",
        )


"""
    @mock.patch("hydrofunctions.validate.check_parameter_string")
    def test_NWIS_init_calls_check_parameter_string(self, mock_cps):

    def test_NWIS_start_defaults_to_None(self):
        actual = station.NWIS('01541200')
        self.assertIsNone(actual.start_date)

    def test_NWIS_end_defaults_to_None(self):
        actual = station.NWIS('01541200')
        self.assertIsNone(actual.end_date)

    def test_NWIS_setters_work(self):
        site = "01582500"
        service = "dv"
        start = "2011-01-01"
        end = "2011-01-02"
        actual = station.NWIS(site, service, start, end)
        self.assertIsInstance(actual, station.NWIS)
        self.assertEqual(actual.site, site)
        self.assertEqual(actual.service, service)
        self.assertEqual(actual.start_date, start)
        self.assertEqual(actual.end_date, end)
        # self.assertIs(type(actual.fetch), function)

    def test_NWIS_setters_parameterCd(self):
        site = "01582500"
        service = "dv"
        start = "2011-01-01"
        end = "2011-01-02"
        parameterCd = "00065"
        actual = station.NWIS(site, service, start, end,
                              parameterCd=parameterCd)
        self.assertIsInstance(actual, station.NWIS)
        self.assertEqual(actual.site, site)
        self.assertEqual(actual.service, service)
        self.assertEqual(actual.start_date, start)
        self.assertEqual(actual.parameterCd, parameterCd)

    @mock.patch("hydrofunctions.hydrofunctions.get_nwis")
    @mock.patch("hydrofunctions.hydrofunctions.get_nwis_property")
    def test_NWIS_get_data_calls_get_nwis_correctly(self, mock_get_prop, mock_get_nwis):
        site = "01582500"
        service = "dv"
        start = "2011-01-01"
        end = "2011-01-02"
        parameterCd = "00060"

        actual = station.NWIS(site, service, start, end, parameterCd=parameterCd)
        try_it_out = actual.get_data()
        # try_it_out should be a response object, I think
        mock_get_nwis.assert_called_once_with(site, service, start, end,
                                              parameterCd=parameterCd,
                                              period=None, stateCd=None,
                                              countyCd=None, bBox=None)

    @mock.patch("hydrofunctions.hydrofunctions.get_nwis")
    @mock.patch("hydrofunctions.hydrofunctions.get_nwis_property")
    def test_NWIS_get_data_calls_get_nwis_mult_sites(self, mock_get_prop, mock_get_nwis):
        site = ["01638500", "01646502"]
        siteEx = "01638500,01646502"
        service = "dv"
        start = "2011-01-01"
        end = "2011-01-02"
        parameterCd = None
        actual = station.NWIS(site, service, start, end)
        try_it_out = actual.get_data()
        # try_it_out should be an instance of NWIS.
        mock_get_nwis.assert_called_once_with(siteEx, service, start, end,
                                              parameterCd='all',
                                              period=None, stateCd=None,
                                              countyCd=None, bBox=None)

    @mock.patch('requests.get')
    @mock.patch("hydrofunctions.hydrofunctions.get_nwis_property")
    def test_NWIS_get_data_accepts_countyCd_array(self, mock_get_prop, mock_get):
        start = "2017-01-01"
        end = "2017-12-31"
        cnty = ['51059', '51061']
        cnty = validate.check_parameter_string(cnty, 'county')
        service2 = 'dv'

        expected_url = 'https://waterservices.usgs.gov/nwis/dv/?'
        expected_headers = {'max-age': '120', 'Accept-encoding': 'gzip'}
        expected_params = {'format': 'json,1.1', 'sites': None,
                           'stateCd': None, 'countyCd': cnty, 'bBox': None,
                           'parameterCd': None, 'period': None,
                           'startDT': start, 'endDT': end}

        expected = fakeResponse(200)

        mock_get.return_value = expected
        actual = station.NWIS(None, service2, start, countyCd=cnty,
                              end_date=end).get_data()
        mock_get.assert_called_once_with(expected_url, params=expected_params,
                                         headers=expected_headers)

    @mock.patch('requests.get')
    @mock.patch("hydrofunctions.hydrofunctions.get_nwis_property")
    def test_NWIS_get_data_accepts_parameterCd_array(self, mock_get_prop, mock_get):
        site = '01541200'
        start = "2017-01-01"
        end = "2017-12-31"
        service = 'iv'
        parameterCd = ['00060','00065']
        expected_parameterCd = '00060,00065'

        expected_url = 'https://waterservices.usgs.gov/nwis/iv/?'
        expected_headers = {'max-age': '120', 'Accept-encoding': 'gzip'}
        expected_params = {'format': 'json,1.1', 'sites': site,
                           'stateCd': None, 'countyCd': None, 'bBox': None,
                           'parameterCd': expected_parameterCd, 'period': None,
                           'startDT': start, 'endDT': end}

        expected = fakeResponse(200)

        mock_get.return_value = expected
        actual = station.NWIS(site, service, start, end_date=end,
                              parameterCd=parameterCd).get_data()
        mock_get.assert_called_once_with(expected_url, params=expected_params,
                                         headers=expected_headers)

    @mock.patch('requests.get')
    @mock.patch("hydrofunctions.hydrofunctions.get_nwis_property")
    def test_NWIS_get_data_converts_parameterCd_all_to_None(self, mock_get_prop, mock_get):
        site = '01541200'
        service = 'iv'
        parameterCd = 'all'
        expected_parameterCd = None
        expected_url = 'https://waterservices.usgs.gov/nwis/' + service + '/?'
        expected_headers = {'max-age': '120', 'Accept-encoding': 'gzip'}
        expected_params = {'format': 'json,1.1', 'sites': site,
                           'stateCd': None, 'countyCd': None, 'bBox': None,
                           'parameterCd': expected_parameterCd, 'period': None,
                           'startDT': None, 'endDT': None}
        expected = fakeResponse(200)
        mock_get.return_value = expected
        actual = station.NWIS(site, service, parameterCd = parameterCd).get_data()
        mock_get.assert_called_once_with(expected_url, params=expected_params,
                                         headers=expected_headers)

    # Now test .df() and .json()
    @unittest.skip("Test this differently")
    def test_NWIS_df_returns_df(self):
        site = "01582500"
        service = "dv"
        start = "2011-01-01"
        end = "2011-01-02"
        actual = station.NWIS(site, service, start, end)
        # You don't need to test the following like this.
        # Just test that actual.df() returns nothing if you call before get_data()
        # And actual.df() returns a df if you call after .get_data()
        actualdf = actual.get_data().df()  # returns a dataframe
        self.assertIs(type(actualdf), pd.core.frame.DataFrame,
                      msg="Did not return a df")

    def test_NWIS_raises_ValueError_too_many_location_arguments(self):
        with self.assertRaises(ValueError):
            station.NWIS('01541000', stateCd='MD')

    def test_NWIS_raises_ValueError_start_and_period_arguments(self):
        with self.assertRaises(ValueError):
            station.NWIS('01541000', start_date='2013-02-02', period='P9D')
    """


if __name__ == "__main__":
    unittest.main(verbosity=2)
